from typing import Union

import pytest

import jwskate.jwa
from jwskate import P_521, ECJwk, InvalidJwe, JweCompact, Jwk, RSAJwk, SymmetricJwk

JWCRYPTO_UNSUPPORTED_ALGS = ["RSA-OAEP-384", "RSA-OAEP-512"]
jwskate.jwa.RsaEsPcks1v1_5.read_only = False  # turn off read only for that alg


def test_jwe() -> None:
    plaintext = b"The true sign of intelligence is not knowledge but imagination."
    alg = "RSA-OAEP"
    enc = "A256GCM"
    cek = bytes.fromhex(
        "b1a1f480548fe1733fb403ff6b9ad4f68a076e5b702e22692f82cb2e7aea40fc"
    )
    jwk = Jwk(
        {
            "kty": "RSA",
            "n": "oahUIoWw0K0usKNuOR6H4wkf4oBUXHTxRvgb48E-BVvxkeDNjbC4he8rUW"
            "cJoZmds2h7M70imEVhRU5djINXtqllXI4DFqcI1DgjT9LewND8MW2Krf3S"
            "psk_ZkoFnilakGygTwpZ3uesH-PFABNIUYpOiN15dsQRkgr0vEhxN92i2a"
            "sbOenSZeyaxziK72UwxrrKoExv6kc5twXTq4h-QChLOln0_mtUZwfsRaMS"
            "tPs6mS6XrgxnxbWhojf663tuEQueGC-FCMfra36C9knDFGzKsNa7LZK2dj"
            "YgyD3JR_MB_4NUJW_TqOQtwHYbxevoJArm-L5StowjzGy-_bq6Gw",
            "e": "AQAB",
            "d": "kLdtIj6GbDks_ApCSTYQtelcNttlKiOyPzMrXHeI-yk1F7-kpDxY4-WY5N"
            "WV5KntaEeXS1j82E375xxhWMHXyvjYecPT9fpwR_M9gV8n9Hrh2anTpTD9"
            "3Dt62ypW3yDsJzBnTnrYu1iwWRgBKrEYY46qAZIrA2xAwnm2X7uGR1hghk"
            "qDp0Vqj3kbSCz1XyfCs6_LehBwtxHIyh8Ripy40p24moOAbgxVw3rxT_vl"
            "t3UVe4WO3JkJOzlpUf-KTVI2Ptgm-dARxTEtE-id-4OJr0h-K-VFs3VSnd"
            "VTIznSxfyrj8ILL6MG_Uv8YAu7VILSB3lOW085-4qE3DzgrTjgyQ",
            "p": "1r52Xk46c-LsfB5P442p7atdPUrxQSy4mti_tZI3Mgf2EuFVbUoDBvaRQ-"
            "SWxkbkmoEzL7JXroSBjSrK3YIQgYdMgyAEPTPjXv_hI2_1eTSPVZfzL0lf"
            "fNn03IXqWF5MDFuoUYE0hzb2vhrlN_rKrbfDIwUbTrjjgieRbwC6Cl0",
            "q": "wLb35x7hmQWZsWJmB_vle87ihgZ19S8lBEROLIsZG4ayZVe9Hi9gDVCOBm"
            "UDdaDYVTSNx_8Fyw1YYa9XGrGnDew00J28cRUoeBB_jKI1oma0Orv1T9aX"
            "IWxKwd4gvxFImOWr3QRL9KEBRzk2RatUBnmDZJTIAfwTs0g68UZHvtc",
            "dp": "ZK-YwE7diUh0qR1tR7w8WHtolDx3MZ_OTowiFvgfeQ3SiresXjm9gZ5KL"
            "hMXvo-uz-KUJWDxS5pFQ_M0evdo1dKiRTjVw_x4NyqyXPM5nULPkcpU827"
            "rnpZzAJKpdhWAgqrXGKAECQH0Xt4taznjnd_zVpAmZZq60WPMBMfKcuE",
            "dq": "Dq0gfgJ1DdFGXiLvQEZnuKEN0UUmsJBxkjydc3j4ZYdBiMRAy86x0vHCj"
            "ywcMlYYg4yoC4YZa9hNVcsjqA3FeiL19rk8g6Qn29Tt0cj8qqyFpz9vNDB"
            "UfCAiJVeESOjJDZPYHdHY8v1b-o-Z2X5tvLx-TCekf7oxyeKDUqKWjis",
            "qi": "VIMpMYbPf47dT1w_zDUXfPimsSegnMOA1zTaX7aGk_8urY6R8-ZW1FxU7"
            "AlWAyLWybqq6t16VFd7hQd0y6flUK4SlOydB61gwanOsXGOAOv82cHq0E3"
            "eL4HrtZkUuKvnPrMnsUUFlfUdybVzxyjz9JF_XyaY14ardLSjf4L_FNY",
        }
    )
    iv = bytes.fromhex("e3c575fc02dbe944b4e14ddb")

    jwe = JweCompact.encrypt(plaintext, jwk, alg=alg, enc=enc, cek=cek, iv=iv)

    assert jwe.initialization_vector == bytes.fromhex("e3c575fc02dbe944b4e14ddb")

    assert jwe.alg == alg
    assert jwe.enc == enc
    assert jwe.decrypt(jwk) == plaintext

    assert jwe.get_header("foo") is None


def test_jwe_decrypt() -> None:
    jwe = (
        "eyJhbGciOiJSU0EtT0FFUCIsImVuYyI6IkEyNTZHQ00ifQ."
        "OKOawDo13gRp2ojaHV7LFpZcgV7T6DVZKTyKOMTYUmKoTCVJRgckCL9kiMT03JGe"
        "ipsEdY3mx_etLbbWSrFr05kLzcSr4qKAq7YN7e9jwQRb23nfa6c9d-StnImGyFDb"
        "Sv04uVuxIp5Zms1gNxKKK2Da14B8S4rzVRltdYwam_lDp5XnZAYpQdb76FdIKLaV"
        "mqgfwX7XWRxv2322i-vDxRfqNzo_tETKzpVLzfiwQyeyPGLBIO56YJ7eObdv0je8"
        "1860ppamavo35UgoRdbYaBcoh9QcfylQr66oc6vFWXRcZ_ZT2LawVCWTIy3brGPi"
        "6UklfCpIMfIjf7iGdXKHzg."
        "48V1_ALb6US04U3b."
        "5eym8TW_c8SuK0ltJ3rpYIzOeDQz7TALvtu6UG9oMo4vpzs9tX_EFShS8iB7j6ji"
        "SdiwkIr3ajwQzaBtQD_A."
        "XFBoMYUZodetZdvTiFvSkQ"
    )

    jwk = Jwk(
        {
            "kty": "RSA",
            "n": "oahUIoWw0K0usKNuOR6H4wkf4oBUXHTxRvgb48E-BVvxkeDNjbC4he8rUW"
            "cJoZmds2h7M70imEVhRU5djINXtqllXI4DFqcI1DgjT9LewND8MW2Krf3S"
            "psk_ZkoFnilakGygTwpZ3uesH-PFABNIUYpOiN15dsQRkgr0vEhxN92i2a"
            "sbOenSZeyaxziK72UwxrrKoExv6kc5twXTq4h-QChLOln0_mtUZwfsRaMS"
            "tPs6mS6XrgxnxbWhojf663tuEQueGC-FCMfra36C9knDFGzKsNa7LZK2dj"
            "YgyD3JR_MB_4NUJW_TqOQtwHYbxevoJArm-L5StowjzGy-_bq6Gw",
            "e": "AQAB",
            "d": "kLdtIj6GbDks_ApCSTYQtelcNttlKiOyPzMrXHeI-yk1F7-kpDxY4-WY5N"
            "WV5KntaEeXS1j82E375xxhWMHXyvjYecPT9fpwR_M9gV8n9Hrh2anTpTD9"
            "3Dt62ypW3yDsJzBnTnrYu1iwWRgBKrEYY46qAZIrA2xAwnm2X7uGR1hghk"
            "qDp0Vqj3kbSCz1XyfCs6_LehBwtxHIyh8Ripy40p24moOAbgxVw3rxT_vl"
            "t3UVe4WO3JkJOzlpUf-KTVI2Ptgm-dARxTEtE-id-4OJr0h-K-VFs3VSnd"
            "VTIznSxfyrj8ILL6MG_Uv8YAu7VILSB3lOW085-4qE3DzgrTjgyQ",
            "p": "1r52Xk46c-LsfB5P442p7atdPUrxQSy4mti_tZI3Mgf2EuFVbUoDBvaRQ-"
            "SWxkbkmoEzL7JXroSBjSrK3YIQgYdMgyAEPTPjXv_hI2_1eTSPVZfzL0lf"
            "fNn03IXqWF5MDFuoUYE0hzb2vhrlN_rKrbfDIwUbTrjjgieRbwC6Cl0",
            "q": "wLb35x7hmQWZsWJmB_vle87ihgZ19S8lBEROLIsZG4ayZVe9Hi9gDVCOBm"
            "UDdaDYVTSNx_8Fyw1YYa9XGrGnDew00J28cRUoeBB_jKI1oma0Orv1T9aX"
            "IWxKwd4gvxFImOWr3QRL9KEBRzk2RatUBnmDZJTIAfwTs0g68UZHvtc",
            "dp": "ZK-YwE7diUh0qR1tR7w8WHtolDx3MZ_OTowiFvgfeQ3SiresXjm9gZ5KL"
            "hMXvo-uz-KUJWDxS5pFQ_M0evdo1dKiRTjVw_x4NyqyXPM5nULPkcpU827"
            "rnpZzAJKpdhWAgqrXGKAECQH0Xt4taznjnd_zVpAmZZq60WPMBMfKcuE",
            "dq": "Dq0gfgJ1DdFGXiLvQEZnuKEN0UUmsJBxkjydc3j4ZYdBiMRAy86x0vHCj"
            "ywcMlYYg4yoC4YZa9hNVcsjqA3FeiL19rk8g6Qn29Tt0cj8qqyFpz9vNDB"
            "UfCAiJVeESOjJDZPYHdHY8v1b-o-Z2X5tvLx-TCekf7oxyeKDUqKWjis",
            "qi": "VIMpMYbPf47dT1w_zDUXfPimsSegnMOA1zTaX7aGk_8urY6R8-ZW1FxU7"
            "AlWAyLWybqq6t16VFd7hQd0y6flUK4SlOydB61gwanOsXGOAOv82cHq0E3"
            "eL4HrtZkUuKvnPrMnsUUFlfUdybVzxyjz9JF_XyaY14ardLSjf4L_FNY",
        }
    )

    plaintext = b"The true sign of intelligence is not knowledge but imagination."
    alg = "RSA-OAEP"
    enc = "A256GCM"

    j = JweCompact(jwe)
    assert j.alg == alg
    assert j.enc == enc
    assert JweCompact(jwe).decrypt(jwk) == plaintext

    assert str(JweCompact(jwe)) == jwe
    assert bytes(JweCompact(jwe)) == jwe.encode()


def test_invalid_jwe() -> None:
    with pytest.raises(InvalidJwe):
        JweCompact("foo")
    with pytest.raises(InvalidJwe):
        JweCompact("foo!.foo!.foo!.foo!")
    with pytest.raises(InvalidJwe):
        JweCompact("eyJhbGciOiJSU0EtT0FFUCIsImVuYyI6IkEyNTZHQ00ifQ.foo!.foo!.foo!")
    with pytest.raises(InvalidJwe):
        JweCompact(
            "eyJhbGciOiJSU0EtT0FFUCIsImVuYyI6IkEyNTZHQ00ifQ."
            "OKOawDo13gRp2ojaHV7LFpZcgV7T6DVZKTyKOMTYUmKoTCVJRgckCL9kiMT03JGe"
            "ipsEdY3mx_etLbbWSrFr05kLzcSr4qKAq7YN7e9jwQRb23nfa6c9d-StnImGyFDb"
            "Sv04uVuxIp5Zms1gNxKKK2Da14B8S4rzVRltdYwam_lDp5XnZAYpQdb76FdIKLaV"
            "mqgfwX7XWRxv2322i-vDxRfqNzo_tETKzpVLzfiwQyeyPGLBIO56YJ7eObdv0je8"
            "1860ppamavo35UgoRdbYaBcoh9QcfylQr66oc6vFWXRcZ_ZT2LawVCWTIy3brGPi"
            "6UklfCpIMfIjf7iGdXKHzg."
            "foo!.foo!"
        )
    with pytest.raises(InvalidJwe):
        JweCompact(
            "eyJhbGciOiJSU0EtT0FFUCIsImVuYyI6IkEyNTZHQ00ifQ."
            "OKOawDo13gRp2ojaHV7LFpZcgV7T6DVZKTyKOMTYUmKoTCVJRgckCL9kiMT03JGe"
            "ipsEdY3mx_etLbbWSrFr05kLzcSr4qKAq7YN7e9jwQRb23nfa6c9d-StnImGyFDb"
            "Sv04uVuxIp5Zms1gNxKKK2Da14B8S4rzVRltdYwam_lDp5XnZAYpQdb76FdIKLaV"
            "mqgfwX7XWRxv2322i-vDxRfqNzo_tETKzpVLzfiwQyeyPGLBIO56YJ7eObdv0je8"
            "1860ppamavo35UgoRdbYaBcoh9QcfylQr66oc6vFWXRcZ_ZT2LawVCWTIy3brGPi"
            "6UklfCpIMfIjf7iGdXKHzg."
            "48V1_ALb6US04U3b."
            "5eym8TW_c8SuK0ltJ3rpYIzOeDQz7TALvtu6UG9oMo4vpzs9tX_EFShS8iB7j6ji"
            "SdiwkIr3ajwQzaBtQD_A."
            ".foo!"
        )


EC_P521_PRIVATE_KEY = {
    "kty": "EC",
    "kid": "bilbo.baggins@hobbiton.example",
    "use": "enc",
    "crv": "P-521",
    "x": "AHKZLLOsCOzz5cY97ewNUajB957y-C-U88c3v13nmGZx6sYl_oJXu9"
    "A5RkTKqjqvjyekWF-7ytDyRXYgCF5cj0Kt",
    "y": "AdymlHvOiLxXkEhayXQnNCvDX4h9htZaCJN34kfmC6pV5OhQHiraVy"
    "SsUdaQkAgDPrwQrJmbnX9cwlGfP-HqHZR1",
    "d": "AAhRON2r9cqXX1hg-RoI6R1tX5p2rUAYdmpHZoC1XNM56KtscrX6zb"
    "KipQrCW9CGZH3T4ubpnoTKLDYJ_fF3_rJt",
}

RSA_PRIVATE_KEY = {
    "kty": "RSA",
    "kid": "bilbo.baggins@hobbiton.example",
    "use": "enc",
    "n": "n4EPtAOCc9AlkeQHPzHStgAbgs7bTZLwUBZdR8_KuKPEHLd4rHVTeT"
    "-O-XV2jRojdNhxJWTDvNd7nqQ0VEiZQHz_AJmSCpMaJMRBSFKrKb2wqV"
    "wGU_NsYOYL-QtiWN2lbzcEe6XC0dApr5ydQLrHqkHHig3RBordaZ6Aj-"
    "oBHqFEHYpPe7Tpe-OfVfHd1E6cS6M1FZcD1NNLYD5lFHpPI9bTwJlsde"
    "3uhGqC0ZCuEHg8lhzwOHrtIQbS0FVbb9k3-tVTU4fg_3L_vniUFAKwuC"
    "LqKnS2BYwdq_mzSnbLY7h_qixoR7jig3__kRhuaxwUkRz5iaiQkqgc5g"
    "HdrNP5zw",
    "e": "AQAB",
    "d": "bWUC9B-EFRIo8kpGfh0ZuyGPvMNKvYWNtB_ikiH9k20eT-O1q_I78e"
    "iZkpXxXQ0UTEs2LsNRS-8uJbvQ-A1irkwMSMkK1J3XTGgdrhCku9gRld"
    "Y7sNA_AKZGh-Q661_42rINLRCe8W-nZ34ui_qOfkLnK9QWDDqpaIsA-b"
    "MwWWSDFu2MUBYwkHTMEzLYGqOe04noqeq1hExBTHBOBdkMXiuFhUq1BU"
    "6l-DqEiWxqg82sXt2h-LMnT3046AOYJoRioz75tSUQfGCshWTBnP5uDj"
    "d18kKhyv07lhfSJdrPdM5Plyl21hsFf4L_mHCuoFau7gdsPfHPxxjVOc"
    "OpBrQzwQ",
    "p": "3Slxg_DwTXJcb6095RoXygQCAZ5RnAvZlno1yhHtnUex_fp7AZ_9nR"
    "aO7HX_-SFfGQeutao2TDjDAWU4Vupk8rw9JR0AzZ0N2fvuIAmr_WCsmG"
    "peNqQnev1T7IyEsnh8UMt-n5CafhkikzhEsrmndH6LxOrvRJlsPp6Zv8"
    "bUq0k",
    "q": "uKE2dh-cTf6ERF4k4e_jy78GfPYUIaUyoSSJuBzp3Cubk3OCqs6grT"
    "8bR_cu0Dm1MZwWmtdqDyI95HrUeq3MP15vMMON8lHTeZu2lmKvwqW7an"
    "V5UzhM1iZ7z4yMkuUwFWoBvyY898EXvRD-hdqRxHlSqAZ192zB3pVFJ0"
    "s7pFc",
    "dp": "B8PVvXkvJrj2L-GYQ7v3y9r6Kw5g9SahXBwsWUzp19TVlgI-YV85q"
    "1NIb1rxQtD-IsXXR3-TanevuRPRt5OBOdiMGQp8pbt26gljYfKU_E9xn"
    "-RULHz0-ed9E9gXLKD4VGngpz-PfQ_q29pk5xWHoJp009Qf1HvChixRX"
    "59ehik",
    "dq": "CLDmDGduhylc9o7r84rEUVn7pzQ6PF83Y-iBZx5NT-TpnOZKF1pEr"
    "AMVeKzFEl41DlHHqqBLSM0W1sOFbwTxYWZDm6sI6og5iTbwQGIC3gnJK"
    "bi_7k_vJgGHwHxgPaX2PnvP-zyEkDERuf-ry4c_Z11Cq9AqC2yeL6kdK"
    "T1cYF8",
    "qi": "3PiqvXQN0zwMeE-sBvZgi289XP9XCQF3VWqPzMKnIgQp7_Tugo6-N"
    "ZBKCQsMf3HaEGBjTVJs_jcK8-TRXvaKe-7ZMaQj8VfBdYkssbu0NKDDh"
    "jJ-GtiseaDVWt7dcH0cfwxgFUHpQh7FoCrjFJ6h6ZEpMF6xmujs4qMpP"
    "z8aaI4",
}

SYMMETRIC_ENCRYPTION_KEY = {
    "kty": "oct",
    "kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037",
    "use": "enc",
    "alg": "A256KW",
    "k": "hJtXIZ2uSN5kbQfbtTNWbpdmhkV8FJG-Onbc6mxCcYg",
}


@pytest.fixture(scope="module")
def ec_p521_private_jwk() -> Jwk:
    """This is the key from [https://datatracker.ietf.org/doc/html/rfc7520#section-3.2]."""
    jwk = Jwk(EC_P521_PRIVATE_KEY)
    assert isinstance(jwk, ECJwk)
    assert jwk.is_private
    assert jwk.kty == "EC"
    assert jwk.kid == "bilbo.baggins@hobbiton.example"
    assert jwk.use == "enc"
    assert jwk.curve == P_521
    assert (
        jwk.x_coordinate
        == 1536512509633812701046363966946458604394346818697394258819956002474017850080242973018354677969345705661882653180474980600600249393774872942916765721086083757
    )
    assert (
        jwk.y_coordinate
        == 6390841077912737365653019441492828723058308087943275742908661475903115398365573971688270351422921162967673593344809147696590449694148618103053589689371825269
    )
    assert (
        jwk.ecc_private_key
        == 111516411687364059110290785888309499157003632541675704481981366909658327544236421609398595734137119337357107861242367481230550517706674527460992238934209133
    )
    return jwk


@pytest.fixture(scope="module")
def ec_p384_private_jwk() -> Jwk:
    return Jwk.generate_for_kty("EC", crv="P-384")


@pytest.fixture(scope="module")
def ec_p256_private_jwk() -> Jwk:
    return ECJwk.generate("P-256")


@pytest.fixture(scope="module")
def okp_x25519_private_jwk() -> Jwk:
    return Jwk.generate_for_kty("OKP", crv="X25519")


@pytest.fixture(scope="module")
def okp_x448_private_jwk() -> Jwk:
    return Jwk.generate_for_kty("OKP", crv="X448")


@pytest.fixture(scope="module")
def rsa_private_jwk() -> Jwk:
    """This is the key from [https://datatracker.ietf.org/doc/html/rfc7520#section-3.4]."""
    jwk = Jwk(RSA_PRIVATE_KEY)
    assert isinstance(jwk, RSAJwk)
    assert jwk.is_private
    assert jwk.kty == "RSA"
    assert jwk.kid == "bilbo.baggins@hobbiton.example"
    assert jwk.use == "enc"
    assert (
        jwk.modulus
        == 20135533008613362683983973718862990570890949482783547491074937566048838943004157274484500282679051238967930814182837332509745335321694730867914487474360313056717004122048241683576190451001206594369003880452220552186311851010130037332999299892700953157894377718386086768938058299374235398748350321163975673243254998238224668780038242796491971359194173117243083075284176788910883569789455869367283514387223800602948314723218768921623931105285092074647944930960873919066358313244754717122611255711161319444897038896496343014060976689972635082113758993979473481511550731351324005908071126862383299605264835961097030990287
    )
    assert jwk.exponent == 65537
    assert (
        jwk.private_exponent
        == 13809785886921180797407749068700942981528089435771470964933339849531763931979658226246689941649114165877756904281400924998825599768673188627050679509247407590724566295036763464811978094688530662125180988122227281988347728446577123121956338123153675672384095849459356212372062261627372823030362241721140402534827920608159114397698659666564432536888691574732214992533122683199444713216582417478411683595580959548347164585150169575401765252679902902625147728179779840007171037544685861186766305472768461620252400817122383292377247352244817836750259386865218586191336825522033117854784997616980884291365145982447543137217
    )
    assert (
        jwk.first_prime_factor
        == 155305159528675998315587554014523516083078608902053750652196202749677048642358299363759977556059891120561885641903854876605754500853726315968216134808579359395711784936178882676585818087673577574114127693348589707935410407050296794012012146664933439273320539907415872853360575628122941817407358922423140657993
    )
    assert (
        jwk.second_prime_factor
        == 129651410614568017951696388521026752738348674639303464401419464668770635900253174384526773731502315063946718183900420330879494363511129117909409612613133053606973655487402983938837056343590378935691659259752059752852280269661044647292328198275965901381648329420292300429253481090448036576608977166562458575959
    )
    assert (
        jwk.first_factor_crt_exponent
        == 5452754506240497308759433019323719585992094222860760795439270374399299463296556858507586681668276080205271813815413736836746282351411023590382002932611213604063854267637328669893708400136364221707380683009522939987478513612504889192694812845812469959990700860994002219137017021229395442602718050574418216489
    )
    assert (
        jwk.second_factor_crt_exponent
        == 6103034953475782159405883067387392346274709877813314428160871698478072108156934906636939392429995910283894984471867749832357906395346648896814450690625379104589983172538233447538219918824119490007305320907809395266022012480039103665056876141436971569684073061958920103517814199063615863387684736238234656863
    )
    assert (
        jwk.first_crt_coefficient
        == 155171362656114787674005338316026300971092335901511555016027916093530558354739494408751451514346305783601055276983183518496184496218932854791325892306914322459078533545141895619742270863660955772717492371592287055310307433879513025000807527757356866299049054535097378049521944951691464480088870568286553270414
    )
    return jwk


@pytest.fixture(scope="module")
def symmetric_128_encryption_jwk() -> Jwk:
    return Jwk.generate_for_kty("oct", key_size=128)


@pytest.fixture(scope="module")
def symmetric_192_encryption_jwk() -> Jwk:
    return Jwk.generate_for_kty("oct", key_size=192)


@pytest.fixture(scope="module")
def symmetric_256_encryption_jwk() -> Jwk:
    """This is the key from [https://datatracker.ietf.org/doc/html/rfc7520#section-3.5]."""
    jwk = Jwk(SYMMETRIC_ENCRYPTION_KEY)
    assert isinstance(jwk, SymmetricJwk)
    assert jwk.is_private
    assert jwk.kty == "oct"
    assert jwk.kid == "018c0ae5-4d9b-471b-bfd6-eef314bc7037"
    assert jwk.use == "enc"
    assert jwk.alg == "A256KW"
    assert (
        jwk.key.hex()
        == "849b57219dae48de646d07dbb533566e976686457c1491be3a76dcea6c427188"
    )
    assert jwk.key_size == 32 * 8
    return jwk


@pytest.fixture(scope="module")
def symmetric_384_encryption_jwk() -> Jwk:
    return Jwk.generate_for_kty("oct", key_size=384)


@pytest.fixture(scope="module")
def symmetric_512_encryption_jwk() -> Jwk:
    return Jwk.generate_for_kty("oct", key_size=512)


@pytest.fixture(scope="module")
def encryption_plaintext() -> bytes:
    """This is the plaintext from [https://datatracker.ietf.org/doc/html/rfc7520#section-4]."""
    return (
        "It’s a dangerous business, Frodo, going out your door. "
        "You step onto the road, and if you don't keep your feet, "
        "there’s no knowing where you might be swept off to."
    ).encode()


@pytest.fixture(
    scope="module",
    params=[
        "RSA1_5",
        "RSA-OAEP",
        "RSA-OAEP-256",
        "RSA-OAEP-384",
        "RSA-OAEP-512",
        "A128KW",
        "A192KW",
        "A256KW",
        "dir",
        "ECDH-ES",
        "ECDH-ES+A128KW",
        "ECDH-ES+A192KW",
        "ECDH-ES+A256KW",
        "A128GCMKW",
        "A192GCMKW",
        "A256GCMKW",
        "PBES2-HS256+A128KW",
        "PBES2-HS384+A192KW",
        "PBES2-HS512+A256KW",
    ],
)
def key_management_alg(request: pytest.FixtureRequest) -> str:
    return request.param  # type: ignore[attr-defined,no-any-return]


@pytest.fixture(
    scope="module",
    params=[
        "A128CBC-HS256",
        "A192CBC-HS384",
        "A256CBC-HS512",
        "A128GCM",
        "A192GCM",
        "A256GCM",
    ],
)
def encryption_alg(request: pytest.FixtureRequest) -> str:
    alg: str = request.param  # type: ignore[attr-defined]
    if alg in SymmetricJwk.ENCRYPTION_ALGORITHMS:
        return alg
    pytest.skip(f"Encryption alg {alg} is not supported yet!")


@pytest.fixture(
    scope="module",
    params=[
        'pAvkcJv!$N8HtIuf3W@KaF&2Gv"EAD/BK[_FEoLIuvMS*aG0tm4,.?'.encode(),
    ],
)
def password(request: pytest.FixtureRequest) -> bytes:
    return request.param  # type: ignore


@pytest.fixture(scope="module")
def decryption_jwk(
    key_management_alg: str,
    encryption_alg: str,
    rsa_private_jwk: Jwk,
    ec_p256_private_jwk: Jwk,
    ec_p384_private_jwk: Jwk,
    ec_p521_private_jwk: Jwk,
    okp_x25519_private_jwk: Jwk,
    okp_x448_private_jwk: Jwk,
    symmetric_128_encryption_jwk: Jwk,
    symmetric_192_encryption_jwk: Jwk,
    symmetric_256_encryption_jwk: Jwk,
    symmetric_384_encryption_jwk: Jwk,
    symmetric_512_encryption_jwk: Jwk,
    password: bytes,
) -> Union[Jwk, bytes]:
    if key_management_alg == "dir":
        for key in (
            symmetric_128_encryption_jwk,
            symmetric_192_encryption_jwk,
            symmetric_256_encryption_jwk,
            symmetric_384_encryption_jwk,
            symmetric_512_encryption_jwk,
        ):
            if encryption_alg in key.supported_encryption_algorithms():
                return key
    elif key_management_alg in (
        "PBES2-HS256+A128KW",
        "PBES2-HS384+A192KW",
        "PBES2-HS512+A256KW",
    ):
        return password
    else:
        for key in (
            rsa_private_jwk,
            ec_p521_private_jwk,
            ec_p384_private_jwk,
            ec_p256_private_jwk,
            okp_x25519_private_jwk,
            okp_x448_private_jwk,
            symmetric_128_encryption_jwk,
            symmetric_192_encryption_jwk,
            symmetric_256_encryption_jwk,
        ):
            if key_management_alg in key.supported_key_management_algorithms():
                return key

    assert False, f"No key supports this Key Management alg: {key_management_alg}"


@pytest.fixture(scope="module")
def encryption_jwk(decryption_jwk: Union[Jwk, bytes]) -> Union[Jwk, bytes]:
    if isinstance(decryption_jwk, SymmetricJwk):
        return decryption_jwk
    elif isinstance(decryption_jwk, bytes):
        return decryption_jwk

    return decryption_jwk.public_jwk()


@pytest.fixture(scope="module")
def encrypted_jwe(
    encryption_plaintext: bytes,
    encryption_jwk: Union[Jwk, bytes],
    key_management_alg: str,
    encryption_alg: str,
) -> JweCompact:
    if isinstance(encryption_jwk, Jwk):
        jwe = JweCompact.encrypt(
            plaintext=encryption_plaintext,
            jwk=encryption_jwk,
            alg=key_management_alg,
            enc=encryption_alg,
        )
    elif isinstance(encryption_jwk, bytes):
        jwe = JweCompact.encrypt_with_password(
            plaintext=encryption_plaintext,
            password=encryption_jwk,
            alg=key_management_alg,
            enc=encryption_alg,
        )
    else:
        assert False, "Unsupported encryption key type"
    assert isinstance(jwe, JweCompact)
    return jwe


def test_decrypt(
    encryption_plaintext: bytes,
    encrypted_jwe: JweCompact,
    decryption_jwk: Union[Jwk, bytes],
    key_management_alg: str,
    encryption_alg: str,
) -> None:
    assert encrypted_jwe.alg == key_management_alg
    assert encrypted_jwe.enc == encryption_alg
    if isinstance(decryption_jwk, Jwk):
        assert encrypted_jwe.decrypt(decryption_jwk) == encryption_plaintext
    else:
        assert (
            encrypted_jwe.decrypt_with_password(decryption_jwk) == encryption_plaintext
        )


def test_decrypt_by_jwcrypto(
    encryption_plaintext: bytes,
    encrypted_jwe: JweCompact,
    decryption_jwk: Jwk,
    key_management_alg: str,
    encryption_alg: str,
) -> None:
    """This test decrypts tokens generated by `jwskate` using another lib `jwcrypto`.

    Args:
        encryption_plaintext: the expected plaintext
        encrypted_jwe: the Jwe encrypted by jwskate to decrypt
        decryption_jwk: the Jwk containing the decryption key
        key_management_alg: the key management alg
        encryption_alg: the encryption alg
    """
    import jwcrypto.jwe  # type: ignore[import]
    import jwcrypto.jwk  # type: ignore[import]
    from jwcrypto.common import InvalidJWEOperation, json_encode  # type: ignore[import]

    if key_management_alg in JWCRYPTO_UNSUPPORTED_ALGS:
        pytest.skip(f"jwcrypto doesn't support key management alg {key_management_alg}")

    jwe_algs_and_rsa1_5 = jwcrypto.jwe.default_allowed_algs + ["RSA1_5"]

    jwe = jwcrypto.jwe.JWE(algs=jwe_algs_and_rsa1_5)
    jwe.deserialize(str(encrypted_jwe))
    if isinstance(decryption_jwk, Jwk):
        jwk = jwcrypto.jwk.JWK(**decryption_jwk)
        jwe.decrypt(jwk)
    else:
        password = decryption_jwk
        jwe.decrypt(password)

    assert jwe.plaintext == encryption_plaintext


@pytest.fixture()
def jwcrypto_encrypted_jwe(
    encryption_plaintext: bytes,
    encryption_jwk: Jwk,
    key_management_alg: str,
    encryption_alg: str,
) -> str:
    """Encrypt a JWE using `jwcrypto`, to make sure it validates with `jwskate`.

    Args:
        encryption_plaintext: the plaintext to encrypt
        encryption_jwk: the Jwk to use to for encryption
        key_management_alg: the key management alg
        encryption_alg: the encryption alg

    Returns:
        a JWE token
    """
    import jwcrypto.jwe
    import jwcrypto.jwk
    from jwcrypto.common import json_encode

    if key_management_alg in JWCRYPTO_UNSUPPORTED_ALGS:
        pytest.skip(f"jwcrypto doesn't support key management alg {key_management_alg}")

    jwe_algs_and_rsa1_5 = jwcrypto.jwe.default_allowed_algs + ["RSA1_5"]
    jwe = jwcrypto.jwe.JWE(
        encryption_plaintext,
        protected=json_encode({"alg": key_management_alg, "enc": encryption_alg}),
        algs=jwe_algs_and_rsa1_5,
    )

    if isinstance(encryption_jwk, Jwk):
        jwk = jwcrypto.jwk.JWK(**encryption_jwk)
        jwe.add_recipient(jwk)
    else:
        password = encryption_jwk
        jwe.add_recipient(password)
    token: str = jwe.serialize(True)
    return token


def test_decrypt_from_jwcrypto(
    encryption_plaintext: bytes,
    jwcrypto_encrypted_jwe: str,
    decryption_jwk: Jwk,
    key_management_alg: str,
    encryption_alg: str,
) -> None:
    """Check that `jwskate` decrypts tokens encrypted by `jwcrypto`.

    Args:
        encryption_plaintext: the plaintext
        jwcrypto_encrypted_jwe: the JWE to validate
        decryption_jwk: the decryption key
        key_management_alg: the key management alg
        encryption_alg: the encryption alg
    """
    jwe = JweCompact(jwcrypto_encrypted_jwe)
    assert jwe.alg == key_management_alg
    assert jwe.enc == encryption_alg
    try:
        assert jwe.decrypt(decryption_jwk) == encryption_plaintext
    except Exception:
        cek = jwe.unwrap_cek(decryption_jwk)
        assert (
            False
        ), f"Decryption by JWSkate failed for {jwcrypto_encrypted_jwe}, CEK={cek}"



"""This module implements password-based Key Management Algorithms relying on PBES2."""

from typing import Type, Union

from binapy import BinaPy
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf import pbkdf2

from ..base import BaseKeyManagementAlg
from .aeskw import A128KW, A192KW, A256KW, BaseAesKeyWrap


class BasePbes2(BaseKeyManagementAlg):
    """Base class for PBES2 based algorithms."""

    kwalg: Type[BaseAesKeyWrap]
    hash_alg: hashes.HashAlgorithm

    def __init__(self, password: Union[bytes, str]):
        """Initialize this alg with the given password.

        Args:
            password: the encryption/decryption password to use
        """
        if isinstance(password, str):
            password = password.encode("utf-8")
        self.password = password

    def generate_salt(self, size: int = 12) -> BinaPy:
        """Generate a salt that is suitable for use for encryption.

        Args:
          size: size of the generated salt, in bytes

        Returns:
            the generated salt

        Raises:
            ValueError: if the salt is less than 8 bytes long
        """
        if size < 8:
            raise ValueError("salts used for PBES2 must be at least 8 bytes long")
        return BinaPy.random(size)

    def derive(self, *, salt: bytes, count: int) -> BinaPy:
        """Derive the encryption key based on the configured password, the given salt and the number of PBKDF iterations.

        Args:
          salt: the generated salt
          count: number of PBKDF iterations

        Returns:
            the generated encryption/decryption key
        """
        full_salt = self.name.encode() + b"\0" + salt
        pbkdf = pbkdf2.PBKDF2HMAC(
            algorithm=self.hash_alg,
            length=self.kwalg.key_size // 8,
            salt=full_salt,
            iterations=count,
        )
        return BinaPy(pbkdf.derive(self.password))

    def wrap_key(self, plainkey: bytes, *, salt: bytes, count: int) -> BinaPy:
        """Wrap a key using this alg.

        Args:
          plainkey: the key to wrap
          salt: the salt to use
          count: the number of PBKDF iterations

        Returns:
            the wrapped key
        """
        aes_key = self.derive(salt=salt, count=count)
        return BinaPy(self.kwalg(aes_key).wrap_key(plainkey))

    def unwrap_key(self, cipherkey: bytes, *, salt: bytes, count: int) -> BinaPy:
        """Unwrap a key using this alg.

        Args:
          cipherkey: the wrapped key
          salt: the salt to use
          count: the number of PBKDF iterations

        Returns:
            the unwrapped key
        """
        aes_key = self.derive(salt=salt, count=count)
        return BinaPy(self.kwalg(aes_key).unwrap_key(cipherkey))


class Pbes2_HS256_A128KW(BasePbes2):
    """PBES2 with HMAC SHA-256 and "A128KW" wrapping."""

    name = "PBES2-HS256+A128KW"
    description = __doc__
    kwalg = A128KW
    hash_alg = hashes.SHA256()


class Pbes2_HS384_A192KW(BasePbes2):
    """PBES2 with HMAC SHA-384 and "A192KW" wrapping."""

    name = "PBES2-HS384+A192KW"
    description = __doc__
    kwalg = A192KW
    hash_alg = hashes.SHA384()


class Pbes2_HS512_A256KW(BasePbes2):
    """PBES2 with HMAC SHA-512 and "A256KW" wrapping."""

    name = "PBES2-HS512+A256KW"
    description = __doc__
    kwalg = A256KW
    hash_alg = hashes.SHA512()



"""This module contains classes that describe Elliptic Curves as described in RFC7518."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Any, ClassVar, Dict, Tuple, Union

from binapy import BinaPy
from cryptography.hazmat.primitives.asymmetric import ec


@dataclass
class EllipticCurve:
    """A descriptive class for Elliptic Curves.

    Elliptic Curves have a name, a `cryptography.ec.EllipticCurve`, and
    a coordinate size.
    """

    name: str
    """
    Curve name as defined in [IANA JOSE](https://www.iana.org/assignments/jose/jose.xhtml#web-key-elliptic-curve).
    This name will appear in `alg` or `enc` fields in JOSE headers."""

    cryptography_curve: ec.EllipticCurve
    """`cryptography` curve instance."""

    coordinate_size: int
    """Coordinate size, in bytes."""

    instances: ClassVar[Dict[str, EllipticCurve]] = {}
    """Registry of subclasses, in a {name: instance} mapping."""

    def __post_init__(self) -> None:
        """Automatically register subclasses in the instance registry."""
        self.instances[self.name] = self

    def generate(self) -> Tuple[int, int, int]:
        """Generate a new EC key on this curve.

        Returns:
             a tuple of 4 `int`s: `x` and `y` coordinates (public key) and `d` (private key)
        """
        key = ec.generate_private_key(self.cryptography_curve)
        pn = key.private_numbers()  # type: ignore
        x = pn.public_numbers.x
        y = pn.public_numbers.y
        d = pn.private_value
        return x, y, d

    @classmethod
    def get_curve(
        cls, key: Union[ec.EllipticCurvePublicKey, ec.EllipticCurvePrivateKey]
    ) -> "EllipticCurve":
        """Get the appropriate `EllipticCurve` instance for a given `cryptography` `EllipticCurvePublicKey`.

        Args:
          key: an Elliptic Curve private or public key from `cryptography`.

        Returns:
          the appropriate instance of EllipticCurve for the given key.

        Raises:
            NotImplementedError: if the curve is not supported
        """
        for c in cls.instances.values():
            if c.cryptography_curve.name == key.curve.name:
                return c
        raise NotImplementedError(f"Unsupported Curve {key.curve.name}")

    @classmethod
    def get_jwk_parameters(
        cls, key: Union[ec.EllipticCurvePrivateKey, ec.EllipticCurvePublicKey]
    ) -> Dict[str, Any]:
        """Extract all private and public parameters from a given `cryptography` `EllipticCurvePrivateKey` or `EllipticCurvePublicKey`.

        Args:
          key: an Elliptic Curve public or private key from `cryptography`.

        Returns:
          a dict of JWK parameters matching that key

        Raises:
            TypeError: if the provided key is not an EllipticCurvePrivateKey or EllipticCurvePublicKey
        """
        if not isinstance(key, (ec.EllipticCurvePrivateKey, ec.EllipticCurvePublicKey)):
            raise TypeError(
                "A EllipticCurvePrivateKey or a EllipticCurvePublicKey is required."
            )
        crv = cls.get_curve(key)
        public_numbers: ec.EllipticCurvePublicNumbers
        if isinstance(key, ec.EllipticCurvePrivateKey):
            public_numbers = key.public_key().public_numbers()
        elif isinstance(key, ec.EllipticCurvePublicKey):
            public_numbers = key.public_numbers()
        x = BinaPy.from_int(public_numbers.x, crv.coordinate_size).to("b64u").ascii()
        y = BinaPy.from_int(public_numbers.y, crv.coordinate_size).to("b64u").ascii()
        parameters = {"kty": "EC", "crv": crv.name, "x": x, "y": y}
        if isinstance(key, ec.EllipticCurvePrivateKey):
            pn = key.private_numbers()  # type: ignore
            d = (
                BinaPy.from_int(pn.private_value, crv.coordinate_size)
                .to("b64u")
                .ascii()
            )
            parameters["d"] = d
        return parameters


P_256 = EllipticCurve(
    name="P-256",
    cryptography_curve=ec.SECP256R1(),
    coordinate_size=32,
)
"""P-256 curve"""

P_384 = EllipticCurve(
    name="P-384",
    cryptography_curve=ec.SECP384R1(),
    coordinate_size=48,
)
"""P-384 curve"""

P_521 = EllipticCurve(
    name="P-521",
    cryptography_curve=ec.SECP521R1(),
    coordinate_size=66,
)
"""P-521 curve"""

secp256k1 = EllipticCurve(
    name="secp256k1",
    cryptography_curve=ec.SECP256K1(),
    coordinate_size=32,
)
"""[secp256k1 curve](https://www.rfc-editor.org/rfc/rfc8812.html)"""



"""This module implements RSA based Key Management algorithms."""

from typing import Any, Union

from binapy import BinaPy
from cryptography.hazmat.primitives import asymmetric, hashes
from cryptography.hazmat.primitives.asymmetric import padding

from ..base import BaseAsymmetricAlg, BaseKeyManagementAlg


class BaseRsaKeyWrap(
    BaseKeyManagementAlg,
    BaseAsymmetricAlg[asymmetric.rsa.RSAPrivateKey, asymmetric.rsa.RSAPublicKey],
):
    """Base class for RSA Key Wrapping algorithms."""

    padding: Any

    name = "RSA1_5"
    description = "RSAES-PKCS1-v1_5"

    private_key_class = asymmetric.rsa.RSAPrivateKey
    public_key_class = asymmetric.rsa.RSAPublicKey

    def __init__(
        self, key: Union[asymmetric.rsa.RSAPublicKey, asymmetric.rsa.RSAPrivateKey]
    ):
        """Initialize an alg with a given RSA key.

        Args:
            key: the private or public key to use
        """
        self.key = key

    def wrap_key(self, plainkey: bytes) -> BinaPy:
        """Wrap a symmetric key using this algorithm.

        Args:
          plainkey: the symmetric key to wrap

        Returns:
            the wrapped key

        Raises:
            PublicKeyRequired: if this algorithm is initialized with a private key instead of a public key
        """
        if self.read_only:
            raise NotImplementedError
        with self.public_key_required() as key:
            return BinaPy(key.encrypt(plainkey, self.padding))

    def unwrap_key(self, cipherkey: bytes) -> BinaPy:
        """Unwrap a symmetric key with this alg.

        Args:
          cipherkey: the wrapped key

        Returns:
            the unwrapped clear-text key
        Raises:
            PrivateKeyRequired: if this alg is initialized with a public key instead of a private key
        """
        with self.private_key_required() as key:
            return BinaPy(key.decrypt(cipherkey, self.padding))


class RsaEsPcks1v1_5(BaseRsaKeyWrap):  # noqa: D415
    """RSAES-PKCS1-v1_5"""

    name = "RSA1_5"
    description = __doc__
    read_only = True

    padding = padding.PKCS1v15()


class RsaEsOaep(BaseRsaKeyWrap):  # noqa: D415
    """RSAES OAEP using default parameters"""

    name = "RSA-OAEP"
    description = __doc__

    padding = padding.OAEP(
        mgf=padding.MGF1(algorithm=hashes.SHA1()),
        algorithm=hashes.SHA1(),
        label=None,
    )


class RsaEsOaepSha256(BaseRsaKeyWrap):  # noqa: D415
    """RSAES OAEP using SHA-256 and MGF1 with SHA-256"""

    name = "RSA-OAEP-256"
    description = __doc__

    padding = padding.OAEP(
        mgf=padding.MGF1(algorithm=hashes.SHA256()),
        algorithm=hashes.SHA256(),
        label=None,
    )


class RsaEsOaepSha384(BaseRsaKeyWrap):  # noqa: D415
    """RSA-OAEP using SHA-384 and MGF1 with SHA-384"""

    name = "RSA-OAEP-384"
    description = __doc__

    padding = padding.OAEP(
        mgf=padding.MGF1(algorithm=hashes.SHA384()),
        algorithm=hashes.SHA384(),
        label=None,
    )


class RsaEsOaepSha512(BaseRsaKeyWrap):  # noqa: D415
    """RSA-OAEP using SHA-512 and MGF1 with SHA-512"""

    name = "RSA-OAEP-512"
    description = __doc__

    padding = padding.OAEP(
        mgf=padding.MGF1(algorithm=hashes.SHA512()),
        algorithm=hashes.SHA512(),
        label=None,
    )



import pytest

from jwskate import Jwk, RSAJwk


@pytest.fixture(scope="session")
def rsa_private_jwk() -> Jwk:
    d = {
        "kty": "RSA",
        "n": "oRHn4oGv23ylRL3RSsL4p_e6Ywinnj2N2tT5OLe5pEZTg-LFBhjFxcJaB-p1dh6XX47EtSfa-JHffU0o5ZRK2ySyNDtlrFAkOpAHH6U83ayE2QPYGzrFrrvHDa8wIMUWymzxpPwGgKBwZZqtTT6d-iy4Ux3AWV-bUv6Z7WijHnOy7aVzZ4dFERLVf2FaaYXDET7GO4v-oQ5ss_guYdmewN039jxkjz_KrA-0Fyhalf9hL8IHfpdpSlHosrmjORG5y9LkYK0J6zxSBF5ZvLIBK33BTzPPiCMwKLyAcV6qdcAcvV4kthKO0iUKBK4eE8D0N8HcSPvA9F_PpLS_k5F2lw",
        "e": "AQAB",
        "d": "U-uiZ8-uMquU6GYF_-_p4ooeIK9HthjYKiZA255OKRtDNDoY8X5CvTDv-6PbI3n44J7gOorTeiN20DA9mLBU-Cz8dF5mMQtodOLJ82ECf3T9mpx4ImfSy6GmqqiWaNyHbVyp4o41TRtKtIWMuexgHxLhztx3dZlssidZR-r24kwON7_2JUeY-N6hwmKh3DlsmO9KyOAoTwNjyKxCIqbf7WnZ9gnavG_mLUeXeiHhSgASYMTUCCFm0KOhDWgvaddDKDMqcQUYPonaI19fW1eNtXfXRjFWwlGbqOnOo930yl1LG1CawI0rbxmkDoyjTHLDJlY7Go_gpHlP2maQRPiMcQ",
        "p": "0mzP9sbFxU5YxNNLgUEdRQSO-ojqWrzbI02PfQLGyzXumvOh_Qr73OpHStU8CAAcUBaQdRGidsVdb5cq6JG2zvbEEYiX-dCHqTJs8wfktGCL7eV-ZVh7fhJ1sYVBN20yv8aSH63uUPZnJXR1AUyrvRumuerdPxp8X951PESrJd0",
        "q": "w_SPRMeEtbEvsRcNfpmbRbpO368hcaLjB9Bb_IvxvoiI3aMTMZgwLSyx5hpuv6A86R3wFdRkh2JBKCzG4ItirUyTfVRUY9ItSNyMNplHxELA6I4JG0m6Rh-IO3wG8-h-U-NKllG4SCR8mS9Wvhg7eBZh_LXvKSgKLgalZSSUSQM",
        "dp": "ogg6B3u-VJVk04Mk1A3w3PGKq678Twy37bJOuGOH8njAGD9c0D5B_TXF2gDirgJvytflOtBueui1bzVHTDjQPQRVrG6zICGMJSR4Mpg0axUhCvo53w5IYacTS7QhqO2EM5pTcON87Ikgmf3YDz0bzY3aT7Vj0rCxbx0cx2DVLV0",
        "dq": "TwdPzJ5m4FwgbtxsPdW3cIyuCLp503ms9FbM8nKCQaSRBkohkIvfSijPaozYg4Idbqr7S-KH1K4Ety4v2xl754aNqSscidGXH96K0e5JqlZ9tIysEYxPir5m1A62QyJN6IkvaKZ2munUMneMFUhym4Dzbdb2KHQUfvGBPORexX8",
        "qi": "rymn9AZV0TshtAM32YLo2HNzOOXRVLbwMZUjOeViuUVSyPqtkKNYFHKBpg7QxuzGbl6w32xKLKoW7xmlQKsSCMtFyVFYtv5muRNlQMG79xxX2M65MhUesPoe7YMJR0fHSBQ6yDvOOdP35CEnABh7AiIIW_rs3ngyfIOyAm0XuiE",
    }
    jwk = Jwk(d)
    assert jwk.is_private
    assert jwk.kty == "RSA"
    assert type(jwk) == RSAJwk
    assert jwk.n == d["n"]
    assert jwk.e == d["e"]
    assert jwk.d == d["d"]
    assert jwk.p == d["p"]
    assert jwk.q == d["q"]
    assert jwk.dp == d["dp"]
    assert jwk.dq == d["dq"]
    assert jwk.qi == d["qi"]

    assert jwk.thumbprint() == "Qfq9DOLKNRyptzTJBhCFlzccbA0ac7Ag9GVFL11GAfM"

    assert jwk.supported_encryption_algorithms() == []
    return jwk


@pytest.fixture(scope="session")
def rsa_public_jwk(rsa_private_jwk: Jwk) -> Jwk:
    public_jwk = rsa_private_jwk.public_jwk()

    assert not public_jwk.is_private

    assert public_jwk.n == rsa_private_jwk.n
    assert public_jwk.e == rsa_private_jwk.e
    assert public_jwk.get("d") is None
    assert public_jwk.get("p") is None
    assert public_jwk.get("q") is None
    assert public_jwk.get("dp") is None
    assert public_jwk.get("dq") is None
    assert public_jwk.get("qi") is None

    assert public_jwk.thumbprint() == rsa_private_jwk.thumbprint()

    return public_jwk


def test_sign(rsa_private_jwk: Jwk, rsa_public_jwk: Jwk) -> None:
    signature = rsa_private_jwk.sign(b"Hello World!", alg="RS256")
    assert (
        signature.hex()
        == "2eb2d1f5ef9a55403b7d09cca52955feea3ced6b948d311819ec976e4f40cb3cdf9718de38ecc53f"
        "d060e2994fab378cb64ebcecf1a6da1d5983af8b6d53c2830e0a4815863345ac72f9a6e7b6328f56"
        "78c1a3ed89074fa1e0526f261c5d969c0d059db94fedd51a705ae1870ef4c00cf89b5702c62f20fd"
        "1c3f13b94b15e529a9f6d86810788cf7d6d9e1e296d094af934931d6b845d2c93239943ca678b715"
        "310c2019ac1eca39dc1e8e67153342ab5d8d500ee07e438b316a1e6e2cd11191fb2ddf98ae2d9f62"
        "a6d50f74890d429af57946e744dda52f8341014a9bbc1b82bcaeae8d5458d3433140b88d6fc2c46a"
        "f011c2189fdf6adc27b53e2ae90b6207"
        # TODO: check this value
    )

    assert rsa_public_jwk.verify(b"Hello World!", signature, alg="RS256")


def test_public_jwk(rsa_private_jwk: Jwk) -> None:
    public_jwk = {
        key: val for key, val in rsa_private_jwk.items() if key in ("kty", "n", "e")
    }
    jwk = Jwk(public_jwk)
    assert jwk.is_private is False
    assert jwk.kty == "RSA"
    assert type(jwk) == RSAJwk
    assert jwk.n == rsa_private_jwk["n"]
    assert jwk.e == rsa_private_jwk["e"]

    assert jwk.thumbprint() == rsa_private_jwk.thumbprint()


def test_generate() -> None:
    jwk = RSAJwk.generate(kid="myrsakey")
    assert jwk.kty == "RSA"
    assert jwk.kid == "myrsakey"
    assert "n" in jwk
    assert "d" in jwk
    assert "p" in jwk
    assert "q" in jwk
    assert "dp" in jwk
    assert "dq" in jwk
    assert "qi" in jwk

    public_jwk = jwk.public_jwk()
    assert public_jwk.kty == "RSA"
    assert public_jwk.kid == "myrsakey"
    assert "d" not in public_jwk
    assert "p" not in public_jwk
    assert "q" not in public_jwk
    assert "dp" not in public_jwk
    assert "dq" not in public_jwk
    assert "qi" not in public_jwk



"""This module implement the JWS JSON flat and general formats."""

from __future__ import annotations

from typing import Any, Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Union

from backports.cached_property import cached_property
from binapy import BinaPy

from jwskate.jwk.base import BaseJsonDict, Jwk

from .compact import JwsCompact
from .signature import JwsSignature


class JwsJsonFlat(JwsSignature):
    """Represent a JWS with a single signature in JSON flat format."""

    @cached_property
    def payload(self) -> bytes:
        """The JWS payload, decoded.

        Returns:
            The raw JWS payload.
        """
        payload = self.get("payload")
        if payload is None:
            raise AttributeError("This Jws JSON does not contain a 'payload' member")
        return BinaPy(payload).decode_from("b64u")

    @cached_property
    def jws_signature(self) -> JwsSignature:
        """The JWS signature.

        Returns:
            The JWS signature.
        """
        content = {
            "protected": self["protected"],
            "signature": self["signature"],
        }
        header = self.get("header")
        if header:
            content["header"] = self.header
        return JwsSignature(content)

    @classmethod
    def sign(
        cls,
        payload: bytes,
        jwk: Union[Jwk, Dict[str, Any]],
        alg: Optional[str] = None,
        extra_protected_headers: Optional[Mapping[str, Any]] = None,
        header: Optional[Any] = None,
        **kwargs: Any,
    ) -> JwsJsonFlat:
        """Signs a payload into a JWS in JSON flat format.

        Args:
            payload: the data to sign.
            jwk: the key to use
            alg: the signature alg to use
            extra_protected_headers: additional protected headers to include
            header: the unprotected header to include
            **kwargs: extra attributes to include in the JWS

        Returns:
            The JWS with the payload, signature, header and extra claims.
        """
        signature = super().sign(
            payload, jwk, alg, extra_protected_headers, header, **kwargs
        )
        signature["payload"] = BinaPy(payload).to("b64u").ascii()
        return cls(signature)

    def generalize(self) -> JwsJsonGeneral:
        """Create a JWS in JSON general format from this JWS in JSON flat.

        Returns:
            A JwsJsonGeneral with the same payload and signature.
        """
        content = self.copy()
        protected = content.pop("protected")
        header = content.pop("header", None)
        signature = content.pop("signature")
        jws_signature = {"protected": protected, "signature": signature}
        if header is not None:
            jws_signature["header"] = header
        content["signatures"] = [jws_signature]
        return JwsJsonGeneral(content)

    def signed_part(self) -> bytes:
        """The signed part from this JWS, as bytes.

        This is a concatenation of the protected header and the payload, separated by a dot (`.`).

        Returns:
            The signed data part.
        """
        return JwsSignature.assemble_signed_part(self.protected, self.payload)

    def compact(self) -> JwsCompact:
        """Create a JWS in compact format from this JWS JSON.

        Returns:
            A `JwsCompact` with the same payload and signature.
        """
        return JwsCompact.from_parts(self.signed_part(), self.signature)

    def verify_signature(
        self,
        jwk: Union[Jwk, Dict[str, Any]],
        *,
        alg: Optional[str] = None,
        algs: Optional[Iterable[str]] = None,
    ) -> bool:
        """Verify this JWS signature with a given key.

        Args:
            jwk: the key to use to validate this signature.
            alg: the signature alg, if only 1 is allowed.
            algs: the allowed signature algs, if there are several.

        Returns:
            `True` if the signature is verified, `False` otherwise.
        """
        return self.jws_signature.verify(self.payload, jwk, alg=alg, algs=algs)


class JwsJsonGeneral(BaseJsonDict):
    """Represents a JWS in JSON general format (possibly with multiple signatures)."""

    @cached_property
    def payload(self) -> bytes:
        """The raw signed data.

        Returns:
            The signed data.
        """
        payload = self.get("payload")
        if payload is None:
            raise AttributeError("This Jws JSON does not contain a 'payload' member")
        return BinaPy(payload).decode_from("b64u")

    @classmethod
    def sign(
        cls,
        payload: bytes,
        *signature_parameters: Union[
            Tuple[
                Union[Jwk, Mapping[str, Any]],
                str,
                Optional[Mapping[str, Any]],
                Optional[Mapping[str, Any]],
            ],
            Tuple[
                Union[Jwk, Mapping[str, Any]],
                str,
                Optional[Mapping[str, Any]],
            ],
            Tuple[
                Union[Jwk, Mapping[str, Any]],
                str,
            ],
            Union[Jwk, Mapping[str, Any]],
        ],
    ) -> JwsJsonGeneral:
        """Sign a payload with several keys and return the resulting JWS JSON in general format.

        Args:
            payload: the data to sign
            *signature_parameters: each of those parameter can be:
                - a `(jwk, alg, extra_protected_headers, header)` tuple
                - a `(jwk, alg, extra_protected_headers)` tuple,
                - a `(jwk, alg)` tuple,
                - a `jwk`
                with `jwk` being a Jwk key, alg being the signature algorithm to use, extra_protected_headers a mapping of extra protected headers and values to include, and header the raw unprotected header to include in the signature.

        Returns:
            A JwsJsonGeneral with the generated signatures.
        """
        jws = cls({"payload": BinaPy(payload).to("b64u").ascii()})
        for parameters in signature_parameters:
            jws.add_signature(*parameters)
        return jws

    @cached_property
    def signatures(self) -> List[JwsSignature]:
        """The list of `JwsSignature` from this JWS.

        Returns:
            The list of signatures from this JWS.
        """
        signatures = self.get("signatures")
        if signatures is None:
            raise AttributeError("This Jws JSON does not contain a 'signatures' member")
        return [JwsSignature(sig) for sig in signatures]

    def add_signature(
        self,
        jwk: Union[Jwk, Dict[str, Any]],
        alg: Optional[str] = None,
        extra_protected_headers: Optional[Mapping[str, Any]] = None,
        header: Optional[Mapping[str, Any]] = None,
    ) -> JwsJsonGeneral:
        """Add a new signature in this JWS.

        Args:
            jwk: the private key to use
            alg: the signature algorithm
            extra_protected_headers: additional headers to include, as a {key: value} mapping
            header: the raw unprotected header to include in the signature

        Returns:
            the same JWS with the new signature included.
        """
        self.setdefault("signatures", [])
        self["signatures"].append(
            JwsSignature.sign(self.payload, jwk, alg, extra_protected_headers, header)
        )
        return self

    def signed_part(
        self,
        signature_chooser: Callable[
            [List[JwsSignature]], JwsSignature
        ] = lambda sigs: sigs[0],
    ) -> bytes:
        """Return the signed part from a given signature.

        The signed part is a concatenation of the protected header from a specific signature, then the payload,
        separated by a dot (`.`).
        You can select the specific signature with the `signature_chooser` parameter.
        By default, the first signature is selected.

        Args:
            signature_chooser: a callable that takes the list of signatures from this JWS as parameter and returns the choosen signature.

        Returns:
            The raw signed part from the chosen signature.
        """
        signature = signature_chooser(self.signatures)
        return JwsSignature.assemble_signed_part(signature.protected, self.payload)

    def compact(
        self,
        signature_chooser: Callable[
            [List[JwsSignature]], JwsSignature
        ] = lambda sigs: sigs[0],
    ) -> JwsCompact:
        """Create a compact JWS from a specific signature from this JWS.

        Args:
            signature_chooser: a callable that takes the list of signatures from this JWS as parameter and returns the choosen signature.

        Returns:
            A JwsCompact with the payload and the chosen signature from this JWS.
        """
        return JwsCompact.from_parts(self.signed_part(signature_chooser), self.payload)

    def flatten(
        self,
        signature_chooser: Callable[
            [List[JwsSignature]], JwsSignature
        ] = lambda sigs: sigs[0],
    ) -> JwsJsonFlat:
        """Create a JWS in JSON flat format from a specific signature from this JWS.

        Args:
            signature_chooser:  a callable that takes the list of signatures from this JWS as parameter and returns the choosen signature.

        Returns:
            A JwsJsonFlat with the payload and the chosen signature from this JWS.
        """
        signature = signature_chooser(self.signatures)
        return JwsJsonFlat.from_parts(
            payload=self["payload"],
            protected=signature.protected,
            header=signature.header,
            signature=signature.signature,
        )

    def verify_signature(
        self,
        jwk: Union[Jwk, Dict[str, Any]],
        *,
        alg: Optional[str] = None,
        algs: Optional[Iterable[str]] = None,
    ) -> bool:
        """Verify the signatures from this JWS.

        It will try to validate each signature with the given key, and returns `True` if at least one signature verifies.
        Args:
            jwk: the public key to use
            alg: the signature algorithm to use, if only 1 is allowed.
            algs: the allowed signature algorithms, if there are several.

        Returns:
            `True` if any of the signature verifies with the given key, `False` otherwise.
        """
        for signature in self.signatures:
            if signature.verify(self.payload, jwk, alg=alg, algs=algs):
                return True
        return False



"""This module implements HMAC based signature algorithms."""

from typing import Type

from binapy import BinaPy
from cryptography.hazmat.primitives import hashes, hmac

from ..base import BaseSignatureAlg, BaseSymmetricAlg


class BaseHMACSigAlg(BaseSymmetricAlg, BaseSignatureAlg):
    """Base class for HMAC signature algorithms."""

    mac: Type[hmac.HMAC] = hmac.HMAC
    hash_alg: hashes.HashAlgorithm
    min_key_size: int

    def sign(self, data: bytes) -> BinaPy:  # noqa: D102
        if self.read_only:
            raise NotImplementedError
        m = self.mac(self.key, self.hash_alg)
        m.update(data)
        signature = m.finalize()
        return BinaPy(signature)

    def verify(self, data: bytes, signature: bytes) -> bool:  # noqa: D102
        candidate_signature = self.sign(data)
        return candidate_signature == signature


class HS256(BaseHMACSigAlg):  # noqa: D415
    """HMAC using SHA-256"""

    name = "HS256"
    description = __doc__
    hash_alg = hashes.SHA256()
    min_key_size = 256


class HS384(BaseHMACSigAlg):  # noqa: D415
    """HMAC using SHA-384"""

    name = "HS384"
    description = __doc__
    hash_alg = hashes.SHA384()
    min_key_size = 384


class HS512(BaseHMACSigAlg):  # noqa: D415
    """HMAC using SHA-512"""

    name = "HS512"
    description = __doc__
    hash_alg = hashes.SHA512()
    min_key_size = 512


class HS1(BaseHMACSigAlg):  # noqa: D415
    """HMAC using SHA-1"""

    name = "HS1"
    description = __doc__
    read_only = True
    min_key_size = 160
    hash_alg = hashes.SHA1()



import pytest

from jwskate import Jwk, JwkSet


def test_jwks() -> None:
    keys = [Jwk.generate_for_kty("RSA"), Jwk.generate_for_kty("RSA")]
    jwks = JwkSet(keys=keys)
    assert jwks.jwks == keys

    jwk = Jwk.generate_for_kty("EC")
    keys.append(jwk)
    kid = jwks.add_jwk(jwk)
    assert kid == jwk.kid
    assert jwks.jwks == keys

    data = b"this is a test"
    signature = jwk.sign(data, "ES256")

    assert jwks.verify(data, signature, "ES256")

    jwks.remove_jwk(jwk.kid)

    jwks.remove_jwk("foo")

    assert jwks.is_private

    assert not jwks.verify(data, signature, "ES256")


def test_empty_jwks() -> None:
    jwks = JwkSet()
    assert len(jwks) == 0

    generated_jwk = Jwk.generate_for_kty("RSA")

    kid = jwks.add_jwk(generated_jwk)
    jwk = jwks.get_jwk_by_kid(kid)
    assert jwk.pop("kid") == jwk.thumbprint()
    assert jwk == generated_jwk

    with pytest.raises(KeyError):
        jwks.get_jwk_by_kid("foo")



"""This module implements Encrypted JWT token handling."""

from typing import Union

from .base import Jwt


class EncryptedJwt(Jwt):
    """Represent an encrypted JWT."""

    def __init__(self, value: Union[bytes, str]):
        """Initialize an EncryptedJwt based on its serialized value."""
        raise NotImplementedError



"""This module contains enums for the various identifiers used in JWA and JWK.

See [IANA JOSE](https://www.iana.org/assignments/jose/jose.xhtml).
"""


class SignatureAlgs:
    """Identifiers for Signature algorithms."""

    RS256 = "RS256"
    RS384 = "RS384"
    RS512 = "RS512"
    ES256 = "ES256"
    ES384 = "ES384"
    ES512 = "ES512"
    PS256 = "PS256"
    PS384 = "PS384"
    PS512 = "PS512"
    EdDSA = "EdDSA"

    HS256 = "HS256"
    HS384 = "HS384"
    HS512 = "HS512"

    ALL_SYMMETRIC = [HS256, HS384, HS512]
    ALL_ASYMMETRIC = [
        RS256,
        RS384,
        RS512,
        ES256,
        ES384,
        ES512,
        PS256,
        PS384,
        PS512,
        EdDSA,
    ]


class EncryptionAlgs:
    """Identifiers for Encryption algorithms."""

    A128CBC_HS256 = "A128CBC-HS256"
    A192CBC_HS384 = "A192CBC-HS384"
    A256CBC_HS512 = "A256CBC-HS512"
    A128GCM = "A128GCM"
    A192GCM = "A192GCM"
    A256GCM = "A256GCM"

    ALL = [A128CBC_HS256, A192CBC_HS384, A256CBC_HS512, A128GCM, A192GCM, A256GCM]


class KeyManagementAlgs:
    """Identifiers for Key Management algorithms."""

    RSA1_5 = "RSA1_5"
    RSA_OAEP = "RSA-OAEP"
    RSA_OAEP_256 = "RSA-OAEP-256"
    RSA_OAEP_384 = "RSA-OAEP-384"
    RSA_OAEP_512 = "RSA-OAEP-512"
    ECDH_ES = "ECDH-ES"
    ECDH_ES_A128KW = "ECDH-ES+A128KW"
    ECDH_ES_A192KW = "ECDH-ES+A192KW"
    ECDH_ES_A256KW = "ECDH-ES+A256KW"

    A128KW = "A128KW"
    A192KW = "A192KW"
    A256KW = "A256KW"
    A128GCMKW = "A128GCMKW"
    A192GCMKW = "A192GCMKW"
    A256GCMKW = "A256GCMKW"
    dir = "dir"

    PBES2_HS256_A128KW = "PBES2-HS256+A128KW"
    PBES2_HS384_A192KW = "PBES2-HS384+A192KW"
    PBES2_HS512_A256KW = "PBES2-HS512+A256KW"

    ALL_SYMMETRIC = [
        A128KW,
        A192KW,
        A256KW,
        A128GCMKW,
        A192GCMKW,
        A256GCMKW,
        dir,
        PBES2_HS256_A128KW,
        PBES2_HS384_A192KW,
        PBES2_HS512_A256KW,
    ]
    ALL_ASYMMETRIC = [
        RSA1_5,
        RSA_OAEP,
        RSA_OAEP_256,
        RSA_OAEP_384,
        RSA_OAEP_512,
        ECDH_ES,
        ECDH_ES_A128KW,
        ECDH_ES_A192KW,
        ECDH_ES_A256KW,
    ]



"""This module implements Json Web Key Sets (JWKS)."""

from typing import Any, Dict, Iterable, List, Optional, Union

from .alg import UnsupportedAlg
from .base import BaseJsonDict, Jwk


class JwkSet(BaseJsonDict):
    """A set of JWK keys, with methods for easy management of keys.

    A JwkSet is a dict subclass, so you can do anything with a JwkSet
    that you can do with a dict. In addition, it provides a few helpers
    methods to get the keys, add or remove keys, and verify signatures
    using keys from this set.
    """

    def __init__(
        self,
        jwks: Optional[Dict[str, Any]] = None,
        keys: Optional[Iterable[Jwk]] = None,
    ):
        """Initialize a JwkSet. Multiple inputs can be provided.

        - a `dict` from the parsed JSON object representing this JwkSet (in paramter `jwks`)
        - a list of `Jwk` (in parameter `keys`
        - nothing, to initialize an empty JwkSet

        Args:
            jwks: a dict, containing the JwkSet, parsed as a JSON object.
            keys: a list of Jwk, that will be added to this JwkSet
        """
        if jwks is None and keys is None:
            keys = []

        if jwks is not None:
            keys = jwks.pop("keys", [])
            super().__init__(
                jwks
            )  # init the dict with all the dict content that is not keys
        else:
            super().__init__()

        if keys is not None:
            for jwk in keys:
                self.add_jwk(jwk)

    @property
    def jwks(self) -> List[Jwk]:
        """Return the list of keys from this JwkSet, as `Jwk` instances.

        Returns:
            a list of `Jwk`
        """
        return self.get("keys", [])

    def get_jwk_by_kid(self, kid: str) -> Jwk:
        """Return a Jwk from this JwkSet, based on its kid.

        Args:
          kid: the kid of the key to obtain

        Returns:
            the key with the matching Key ID

        Raises:
            KeyError: if no key matches
        """
        jwk = next(filter(lambda jwk: jwk.get("kid") == kid, self.jwks), None)
        if isinstance(jwk, Jwk):
            return jwk
        raise KeyError(kid)

    def __len__(self) -> int:
        """Return the number of Jwk in this JwkSet.

        Returns:
            the number of keys
        """
        return len(self.jwks)

    def add_jwk(
        self,
        jwk: Union[Jwk, Dict[str, Any]],
        kid: Optional[str] = None,
        use: Optional[str] = None,
    ) -> str:
        """Add a Jwk in this JwkSet.

        Args:
          jwk: the Jwk to add (either a `Jwk` instance, or a dict containing the Jwk parameters)
          kid: the kid to use, if `jwk` doesn't contain one
          use: the defined use for the added Jwk

        Returns:
          the kid from the added Jwk (it may be generated if no kid is provided)
        """
        if not isinstance(jwk, Jwk):
            jwk = Jwk(jwk)

        if "keys" not in self:
            self["keys"] = []

        kid = jwk.get("kid", kid)
        if not kid:
            kid = jwk.thumbprint()
        jwk["kid"] = kid
        use = jwk.get("use", use)
        if use:
            jwk["use"] = use
        self.jwks.append(jwk)

        return kid

    def remove_jwk(self, kid: str) -> None:
        """Removes a Jwk from this JwkSet, based on a `kid`.

        Args:
          kid: the `kid` from the key to be removed.

        Raises:
            KeyError: if no key matches
        """
        try:
            jwk = self.get_jwk_by_kid(kid)
            self.jwks.remove(jwk)
        except KeyError:
            pass

    @property
    def is_private(self) -> bool:
        """True if the JwkSet contains at least one private key.

        Returns:
            `True` if this JwkSet contains at least one private key

        """
        return any(key.is_private for key in self.jwks)

    def public_jwks(self) -> "JwkSet":
        """Return another JwkSet with the public keys associated with the current keys.

        Returns:
            a public JwkSet
        """
        return JwkSet(keys=(key.public_jwk() for key in self.jwks))

    def verify(
        self,
        data: bytes,
        signature: bytes,
        alg: Optional[str] = None,
        algs: Optional[Iterable[str]] = None,
        kid: Optional[str] = None,
    ) -> bool:
        """Verify a signature with the key from this key set.

        It implements multiple techniques to avoid trying all keys:
        If a `kid` is provided, only the key with this `kid` will be tried.
        Otherwise, if an `alg` or several `algs` are provided, only keys that are compatible with the supplied `alg` will be tried.
        Otherwise, keys that have use = signature will be tried.
        And if the signature is still not verified at that point, the keys with no specified alg and use will be tried.

        Args:
          data: the signed data to verify
          signature: the signature to verify against the signed data
          alg: alg to verify the signature, if there is only 1
          algs: list of allowed signature algs, if there are several
          kid: the kid of the Jwk that will be used to validate the signature. If no kid is provided, multiple keys
        from this key set may be tried.

        Returns:
          `True` if the signature validates with any of the tried keys, `False` otherwise
        """
        # if a kid is provided, try only the key matching `kid`
        if kid is not None:
            jwk = self.get_jwk_by_kid(kid)
            return jwk.verify(data, signature, alg=alg, algs=algs)

        # if one or several alg are provided, try only the keys that are compatible with one of the provided alg(s)
        if alg:
            for jwk in self.jwks:
                if jwk.get("alg") == alg:
                    if jwk.verify(data, signature, alg=alg):
                        return True

        if algs:
            for jwk in self.jwks:
                alg = jwk.get("alg")
                if alg is not None and alg in algs:
                    if jwk.verify(data, signature, algs=algs):
                        return True

        # if no kid and no alg are provided, try first the keys flagged for signature verification (`"use": "verify"`)
        for jwk in self.jwks:
            if jwk.get("use") == "verify":
                if jwk.verify(data, signature, alg=alg):
                    return True

        # then with the keys that have no defined `use`
        for jwk in self.jwks:
            if jwk.get("use") is None and jwk.get("alg") is None:
                try:
                    if jwk.verify(data, signature, alg=alg):
                        return True
                except UnsupportedAlg:
                    continue

        # no key matches, so consider the signature invalid
        return False



"""This module implements JWK representing Octet Key Pairs from [RFC8037](https://datatracker.ietf.org/doc/rfc8037/)."""

from __future__ import annotations

from typing import Any, Mapping

from backports.cached_property import cached_property
from binapy import BinaPy
from cryptography.hazmat.primitives.asymmetric import ed448, ed25519, x448, x25519
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
    PublicFormat,
)

from jwskate.jwa import X448, X25519, Ed448, Ed25519, EdDsa, OKPCurve

from .. import EcdhEs, EcdhEs_A128KW, EcdhEs_A192KW, EcdhEs_A256KW
from .base import Jwk, JwkParameter


class UnsupportedOKPCurve(KeyError):
    """Raised when an unsupported OKP curve is requested."""


class OKPJwk(Jwk):
    """Represent an OKP Jwk, with `kty=OKP`."""

    KTY = "OKP"

    CRYPTOGRAPHY_KEY_CLASSES = (
        ed25519.Ed25519PrivateKey,
        ed25519.Ed25519PublicKey,
        ed448.Ed448PrivateKey,
        ed448.Ed448PublicKey,
        x25519.X25519PrivateKey,
        x25519.X25519PublicKey,
        x448.X448PrivateKey,
        x448.X448PublicKey,
    )

    PARAMS = {
        "crv": JwkParameter("Curve", is_private=False, is_required=True, kind="name"),
        "x": JwkParameter(
            "Public Key", is_private=False, is_required=True, kind="b64u"
        ),
        "d": JwkParameter(
            "Private Key", is_private=True, is_required=False, kind="b64u"
        ),
    }

    CURVES: Mapping[str, OKPCurve] = {
        curve.name: curve for curve in [Ed25519, Ed448, X448, X25519]
    }

    SIGNATURE_ALGORITHMS = {alg.name: alg for alg in (EdDsa,)}

    KEY_MANAGEMENT_ALGORITHMS = {
        keyalg.name: keyalg
        for keyalg in [EcdhEs, EcdhEs_A128KW, EcdhEs_A192KW, EcdhEs_A256KW]
    }

    @property
    def is_private(self) -> bool:  # noqa: D102
        return "d" in self

    def _validate(self) -> None:
        if not isinstance(self.crv, str) or self.crv not in self.CURVES:
            raise UnsupportedOKPCurve(self.crv)
        super()._validate()

    @classmethod
    def get_curve(cls, crv: str) -> OKPCurve:
        """Get the OKPCurve instance from a curve identifier.

        Args:
          crv: a crv identifier

        Returns:
            the matching OKPCurve instance

        Raises:
            UnsupportedOKPCurve: if the curve is not supported
        """
        curve = cls.CURVES.get(crv)
        if curve is None:
            raise UnsupportedOKPCurve(crv)
        return curve

    @property
    def curve(self) -> OKPCurve:
        """Get the OKPCurve instance for this key.

        Returns:
            the OKPCurve for this key
        """
        return self.get_curve(self.crv)

    @cached_property
    def public_key(self) -> bytes:
        """Get the public key from this Jwk.

        Returns:
            the public key (from param `x`)
        """
        return BinaPy(self.x).decode_from("b64u")

    @cached_property
    def private_key(self) -> bytes:
        """Get the private key from this Jwk.

        Returns:
            the private key (from param `d`)
        """
        return BinaPy(self.d).decode_from("b64u")

    @classmethod
    def from_cryptography_key(cls, cryptography_key: Any, **kwargs: Any) -> OKPJwk:
        """Initialize a OKPJwk from a `cryptography` key.

        Args:
          cryptography_key: a `cryptography` key
          **kwargs: additional members to include in the Jwk

        Returns:
            the matching OKPJwk
        """
        if isinstance(cryptography_key, ed25519.Ed25519PrivateKey):
            priv = cryptography_key.private_bytes(
                encoding=Encoding.Raw,
                format=PrivateFormat.Raw,
                encryption_algorithm=NoEncryption(),
            )
            pub = cryptography_key.public_key().public_bytes(
                encoding=Encoding.Raw, format=PublicFormat.Raw
            )
            return cls.private(
                crv="Ed25519",
                x=pub,
                d=priv,
            )
        elif isinstance(cryptography_key, ed25519.Ed25519PublicKey):
            pub = cryptography_key.public_bytes(
                encoding=Encoding.Raw, format=PublicFormat.Raw
            )
            return cls.public(
                crv="Ed25519",
                x=pub,
            )
        elif isinstance(cryptography_key, ed448.Ed448PrivateKey):
            priv = cryptography_key.private_bytes(
                encoding=Encoding.Raw,
                format=PrivateFormat.Raw,
                encryption_algorithm=NoEncryption(),
            )
            pub = cryptography_key.public_key().public_bytes(
                encoding=Encoding.Raw, format=PublicFormat.Raw
            )
            return cls.private(
                crv="Ed448",
                x=pub,
                d=priv,
            )
        elif isinstance(cryptography_key, ed448.Ed448PublicKey):
            pub = cryptography_key.public_bytes(
                encoding=Encoding.Raw, format=PublicFormat.Raw
            )
            return cls.public(crv="Ed448", x=pub)
        elif isinstance(cryptography_key, x25519.X25519PrivateKey):
            priv = cryptography_key.private_bytes(
                encoding=Encoding.Raw,
                format=PrivateFormat.Raw,
                encryption_algorithm=NoEncryption(),
            )
            pub = cryptography_key.public_key().public_bytes(
                encoding=Encoding.Raw, format=PublicFormat.Raw
            )
            return cls.private(
                crv="X25519",
                x=pub,
                d=priv,
            )
        elif isinstance(cryptography_key, x25519.X25519PublicKey):
            pub = cryptography_key.public_bytes(
                encoding=Encoding.Raw, format=PublicFormat.Raw
            )
            return cls.public(crv="X25519", x=pub)
        elif isinstance(cryptography_key, x448.X448PrivateKey):
            priv = cryptography_key.private_bytes(
                encoding=Encoding.Raw,
                format=PrivateFormat.Raw,
                encryption_algorithm=NoEncryption(),
            )
            pub = cryptography_key.public_key().public_bytes(
                encoding=Encoding.Raw, format=PublicFormat.Raw
            )
            return cls.private(
                crv="X448",
                x=pub,
                d=priv,
            )
        elif isinstance(cryptography_key, x448.X448PublicKey):
            pub = cryptography_key.public_bytes(
                encoding=Encoding.Raw, format=PublicFormat.Raw
            )
            return cls.public(crv="X448", x=pub)
        else:
            raise TypeError(
                "Unsupported key type for OKP. Supported key types are: "
                + ", ".join(kls.__name__ for kls in cls.CRYPTOGRAPHY_KEY_CLASSES)
            )

    def _to_cryptography_key(self) -> Any:
        """Intialize a `cryptography` key based on this Jwk.

        Returns:
            a Ed25519PrivateKey or a Ed25519PublicKey or a Ed448PrivateKey or a Ed448PublicKey based on the current Jwk

        Raises:
            UnsupportedOKPCurve: if this Jwk curve is not supported.
        """
        if self.curve.name == "Ed25519":
            if self.is_private:
                return ed25519.Ed25519PrivateKey.from_private_bytes(self.private_key)
            else:
                return ed25519.Ed25519PublicKey.from_public_bytes(self.public_key)
        elif self.curve.name == "Ed448":
            if self.is_private:
                return ed448.Ed448PrivateKey.from_private_bytes(self.private_key)
            else:
                return ed448.Ed448PublicKey.from_public_bytes(self.public_key)
        elif self.curve.name == "X25519":
            if self.is_private:
                return x25519.X25519PrivateKey.from_private_bytes(self.private_key)
            else:
                return x25519.X25519PublicKey.from_public_bytes(self.public_key)
        elif self.curve.name == "X448":
            if self.is_private:
                return x448.X448PrivateKey.from_private_bytes(self.private_key)
            else:
                return x448.X448PublicKey.from_public_bytes(self.public_key)
        else:
            raise UnsupportedOKPCurve(self.curve)

    @classmethod
    def public(cls, crv: str, x: bytes, **params: Any) -> OKPJwk:
        """Initialize a public OKPJwk based on the provided parameters.

        Args:
          crv: the key curve
          x: the public key
          **params: additional members to include in the Jwk

        Returns:
            the resulting OKPJwk
        """
        return cls(dict(kty="OKP", crv=crv, x=BinaPy(x).to("b64u").ascii(), **params))

    @classmethod
    def private(cls, crv: str, x: bytes, d: bytes, **params: Any) -> OKPJwk:
        """Initialize a private OKPJwk based on the provided parameters.

        Args:
          crv: the OKP curve
          x: the public key
          d: the private key
          **params: additional members to include in the Jwk

        Returns:
            the resulting OKPJwk
        """
        return cls(
            dict(
                kty=cls.KTY,
                crv=crv,
                x=BinaPy(x).to("b64u").ascii(),
                d=BinaPy(d).to("b64u").ascii(),
                **params,
            )
        )

    @classmethod
    def generate(cls, crv: str = "Ed25519", **params: Any) -> OKPJwk:
        """Generate a private OKPJwk on a given curve.

        Args:
          crv: the curve to use
          **params: additional members to include in the Jwk

        Returns:
            the resulting OKPJwk
        """
        curve = cls.get_curve(crv)
        x, d = curve.generate()
        return cls.private(crv=crv, x=x, d=d, **params)



"""This module implements AES-GCM based encryption algorithms."""

from typing import Optional, Tuple

from binapy import BinaPy
from cryptography.hazmat.primitives.ciphers import aead

from ..base import BaseAESEncryptionAlg


class BaseAESGCM(BaseAESEncryptionAlg):
    """Base class for AES-GCM encryption algorithms."""

    iv_size = 96
    tag_size = 16

    def encrypt(
        self, plaintext: bytes, *, iv: bytes, aad: Optional[bytes]
    ) -> Tuple[BinaPy, BinaPy]:
        """Encrypt a plaintext, with the given IV and Additional Authenticated Data.".

        Args:
          plaintext: the data to encrypt
          iv: the IV to use
          aad: Additional Authenticated Data, if any

        Returns:
            a (ciphertext, authentication_tag) tuple

        Raises:
            ValueError: if the IV size is not appropriate
        """
        if len(iv) * 8 != self.iv_size:
            raise ValueError(f"Invalid IV size, must be {self.iv_size} bits")
        ciphertext_with_tag = BinaPy(aead.AESGCM(self.key).encrypt(iv, plaintext, aad))
        ciphertext, tag = ciphertext_with_tag.cut_at(-self.tag_size)
        return ciphertext, tag

    def decrypt(
        self, ciphertext: bytes, *, iv: bytes, auth_tag: bytes, aad: Optional[bytes]
    ) -> BinaPy:
        """Decrypt a ciphertext.

        Args:
          ciphertext: the data to decrypt
          auth_tag: the Authentication Tag
          iv: the Initialization Vector
          aad: the Additional Authentication Tag

        Returns:
            the decrypted data

        Raises:
            ValueError: if the IV size is not appropriate
        """
        if len(iv) * 8 != self.iv_size:
            raise ValueError(f"Invalid IV size, must be {self.iv_size} bits")
        ciphertext_with_tag = ciphertext + auth_tag
        return BinaPy(aead.AESGCM(self.key).decrypt(iv, ciphertext_with_tag, aad))


class A128GCM(BaseAESGCM):
    """AES GCM using 128-bit key."""

    name = "A128GCM"
    description = __doc__
    key_size = 128


class A192GCM(BaseAESGCM):
    """AES GCM using 192-bit key."""

    name = "A192GCM"
    description = __doc__
    key_size = 192


class A256GCM(BaseAESGCM):
    """AES GCM using 256-bit key."""

    name = "A256GCM"
    description = __doc__
    key_size = 256



"""This modules contains the `Jwt` base class."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from binapy import BinaPy

from jwskate.jwk import Jwk

from ..token import BaseCompactToken

if TYPE_CHECKING:
    from jwskate import EncryptedJwt, SignedJwt  # pragma: no cover


class InvalidJwt(ValueError):
    """Raised when an invalid Jwt is parsed."""


class Jwt(BaseCompactToken):
    """Represents a Json Web Token."""

    def __new__(cls, value: Union[bytes, str]):  # type: ignore
        """Allow parsing both Signed and Encrypted JWTs. Returns the appropriate subclass instance.

        Args:
            value: the token value
        """
        if not isinstance(value, bytes):
            value = value.encode("ascii")

        if cls == Jwt:
            if value.count(b".") == 2:
                from .signed import SignedJwt

                return super().__new__(SignedJwt)
            elif value.count(b".") == 3:
                from .encrypted import EncryptedJwt

                return super().__new__(EncryptedJwt)
        return super().__new__(cls)

    @classmethod
    def sign(
        cls,
        claims: Dict[str, Any],
        jwk: Union[Jwk, Dict[str, Any]],
        alg: Optional[str] = None,
        extra_headers: Optional[Dict[str, Any]] = None,
    ) -> "SignedJwt":
        """Sign a JSON payload with a `Jwk` and returns the resulting `SignedJwt`.

        This method cannot generate a token without a signature. If you want to use an unsigned token (with alg=none),
        use `.unprotected()` instead.

        Args:
          claims: the payload to sign
          jwk: the Jwk to use for signing
          alg: the alg to use for signing
          extra_headers: additional headers to include in the Jwt

        Returns:
          the resulting token
        """
        from .signed import SignedJwt

        jwk = Jwk(jwk)

        alg = alg or jwk.get("alg")
        kid = jwk.get("kid")

        if alg is None:
            raise ValueError("a signing alg is required")

        headers = dict(extra_headers or {}, alg=alg)
        if kid:
            headers["kid"] = kid

        headers_part = BinaPy.serialize_to("json", headers).to("b64u")
        claims_part = BinaPy.serialize_to("json", claims).to("b64u")
        signed_value = b".".join((headers_part, claims_part))
        signature = jwk.sign(signed_value, alg=alg).to("b64u")
        return SignedJwt(b".".join((signed_value, signature)))

    @classmethod
    def unprotected(
        cls,
        claims: Dict[str, Any],
        extra_headers: Optional[Dict[str, Any]] = None,
    ) -> "SignedJwt":
        """Generate a JWT that is not signed and not encrypted (with alg=none).

        Args:
          claims: the claims to set in the token.
          extra_headers: additional headers to insert in the token.

        Returns:
            the resulting token
        """
        from .signed import SignedJwt

        headers = dict(extra_headers or {}, alg="none")

        headers_part = BinaPy.serialize_to("json", headers).to("b64u")
        claims_part = BinaPy.serialize_to("json", claims).to("b64u")
        signed_value = b".".join((headers_part, claims_part))
        signature = b""
        return SignedJwt(b".".join((signed_value, signature)))

    @classmethod
    def sign_and_encrypt(
        cls,
        claims: Dict[str, Any],
        sign_jwk: Union[Jwk, Dict[str, Any]],
        sign_alg: Optional[str],
        enc_jwk: Union[Jwk, Dict[str, Any]],
        enc_alg: Optional[str],
        enc: Optional[str],
    ) -> "EncryptedJwt":
        """Sign then encrypt a payload with a `Jwk` and returns the resulting `EncryptedJwt`.

        NOT IMPLEMENTED YET.

        Args:
          claims: the payload to encrypt
          sign_jwk: the Jwk to use for signature
          sign_alg: the alg to use for signature
          enc_jwk: the Jwk to use for encryption
          enc_alg: the alg to use for CEK encryption
          enc: the alg to use for payload encryption

        Returns:
          the resulting JWE token, with signed JWT as payload
        """
        raise NotImplementedError



"""This module contains all Json Web Key (Jwk) related classes and utilities."""

from .base import Jwt
from .encrypted import EncryptedJwt
from .signed import ExpiredJwt, InvalidClaim, InvalidJwt, InvalidSignature, SignedJwt
from .signer import JwtSigner



"""This module exposes the Encryption algorithms that are available in `jwskate`."""

from .aescbchmac import Aes128CbcHmacSha256, Aes192CbcHmacSha384, Aes256CbcHmacSha512
from .aesgcm import A128GCM, A192GCM, A256GCM

__all__ = [
    "Aes128CbcHmacSha256",
    "Aes192CbcHmacSha384",
    "Aes256CbcHmacSha512",
    "A128GCM",
    "A192GCM",
    "A256GCM",
]



"""This module implements [Json Web Key RFC7517](https://tools.ietf.org/html/rfc7517)."""

from .alg import ExpectedAlgRequired, UnsupportedAlg
from .base import InvalidJwk, Jwk
from .ec import ECJwk, UnsupportedEllipticCurve
from .jwks import JwkSet
from .oct import SymmetricJwk
from .okp import OKPJwk, UnsupportedOKPCurve
from .rsa import RSAJwk



"""This module implements the Edwards-curve Digital Signature Algorithm (EdDSA)."""

from typing import Union

from binapy import BinaPy
from cryptography import exceptions
from cryptography.hazmat.primitives.asymmetric import ed448, ed25519

from ..base import BaseAsymmetricAlg, BaseSignatureAlg


class EdDsa(
    BaseAsymmetricAlg[
        Union[ed25519.Ed25519PrivateKey, ed448.Ed448PrivateKey],
        Union[ed25519.Ed25519PublicKey, ed448.Ed448PublicKey],
    ],
    BaseSignatureAlg,
):
    """EdDSA signature algorithms."""

    private_key_class = (ed25519.Ed25519PrivateKey, ed448.Ed448PrivateKey)
    public_key_class = (ed25519.Ed25519PublicKey, ed448.Ed448PublicKey)

    name = "EdDSA"
    description = __doc__

    def sign(self, data: bytes) -> BinaPy:  # noqa: D102
        with self.private_key_required() as key:
            return BinaPy(key.sign(data))

    def verify(self, data: bytes, signature: bytes) -> bool:  # noqa: D102
        with self.public_key_required() as key:
            try:
                key.verify(signature, data)
                return True
            except exceptions.InvalidSignature:
                return False



"""This module implements JWK representing Symmetric keys."""
from typing import Any, List, Optional, Tuple, Union

from binapy import BinaPy

from jwskate.jwa import (
    A128GCM,
    A128GCMKW,
    A128KW,
    A192GCM,
    A192GCMKW,
    A192KW,
    A256GCM,
    A256GCMKW,
    A256KW,
    HS256,
    HS384,
    HS512,
    Aes128CbcHmacSha256,
    Aes192CbcHmacSha384,
    Aes256CbcHmacSha512,
    BaseAESEncryptionAlg,
    DirectKeyUse,
)

from .alg import select_alg
from .base import Jwk, JwkParameter


class SymmetricJwk(Jwk):
    """Implement Symetric keys, with `kty=oct`."""

    KTY = "oct"
    CRYPTOGRAPHY_KEY_CLASSES = (bytes,)

    PARAMS = {
        "k": JwkParameter("Key Value", is_private=True, is_required=True, kind="b64u"),
    }

    SIGNATURE_ALGORITHMS = {sigalg.name: sigalg for sigalg in [HS256, HS384, HS512]}

    KEY_MANAGEMENT_ALGORITHMS = {
        keyalg.name: keyalg
        for keyalg in [
            A128KW,
            A192KW,
            A256KW,
            A128GCMKW,
            A192GCMKW,
            A256GCMKW,
            DirectKeyUse,
        ]
    }

    ENCRYPTION_ALGORITHMS = {
        keyalg.name: keyalg
        for keyalg in [
            Aes128CbcHmacSha256,
            Aes192CbcHmacSha384,
            Aes256CbcHmacSha512,
            A128GCM,
            A192GCM,
            A256GCM,
        ]
    }

    def public_jwk(self) -> "Jwk":
        """This always raises a ValueError since SymmetricKeys are always private.

        Raises:
            ValueError: symmetric keys are always private, it makes no sense to use them as public keys
        """
        raise ValueError("Symmetric keys don't have a public key")

    @classmethod
    def from_bytes(cls, k: Union[bytes, str], **params: Any) -> "SymmetricJwk":
        """Initializes a SymmetricJwk from a raw secret key. The provided secret key is encoded and used as the `k` parameter for the returned SymetricKey.

        Args:
          k: the key to use
          **params: additional members to include in the Jwk

        Returns:
          the resulting SymmetricJwk
        """
        return cls(dict(kty="oct", k=BinaPy(k).to("b64u").ascii(), **params))

    @classmethod
    def from_cryptography_key(
        cls, cryptography_key: Any, **kwargs: Any
    ) -> "SymmetricJwk":
        """Alias for `from_bytes()` since symmetric keys are simply bytes.

        Args:
            cryptography_key: the key to use
            **kwargs: additional members to include in the Jwk

        Returns:
            the resulting SymmetricJwk
        """
        return cls.from_bytes(cryptography_key, **kwargs)

    @classmethod
    def generate(cls, key_size: int = 128, **params: str) -> "SymmetricJwk":
        """Generate a random SymmetricJwk, with a given key size.

        Args:
          key_size: the size of the generated key, in bits
          **params: additional members to include in the Jwk

        Returns:
            a SymmetricJwk with a randomly generated key
        """
        key = BinaPy.random_bits(key_size)
        return cls.from_bytes(key, **params)

    @classmethod
    def generate_for_alg(cls, alg: str, **params: str) -> "SymmetricJwk":
        """Generate a SymmetricJwk that is suitable for use with the given alg.

        Args:
          alg: the signing algorithm to use this key with
          **params: additional members to include in the Jwk

        Returns:
            the resulting Jwk

        Raises:
            ValueError: if the provided `alg` is not supported
        """
        if alg in cls.SIGNATURE_ALGORITHMS:
            sigalg = cls.SIGNATURE_ALGORITHMS[alg]
            return cls.generate(sigalg.min_key_size, alg=alg, **params)
        if alg in cls.ENCRYPTION_ALGORITHMS:
            encalg = cls.ENCRYPTION_ALGORITHMS[alg]
            return cls.generate(encalg.key_size, alg=alg, **params)
        raise ValueError("Unsupported alg", alg)

    def thumbprint(self, hashalg: str = "SHA256") -> str:
        """Return the key thumbprint as specified by RFC 7638.

        This is reimplemented for SymmetricJwk because the private parameter 'k' must be included.

        Args:
          hashalg: A hash function (defaults to SHA256)

        Returns:
            the calculated thumbprint
        """
        return (
            BinaPy.serialize_to("json", {"k": self.k, "kty": self.kty})
            .to("sha256")
            .to("b64u")
            .ascii()
        )

    def _to_cryptography_key(self) -> BinaPy:
        """Converts this Jwk into a key usable with `cryptography`.

        For SymmetricJwk instances, those are just `bytes` values.

        Returns:
            the raw private key, as `bytes`
        """
        return BinaPy(self.k).decode_from("b64u")

    @property
    def key(self) -> BinaPy:
        """Returns the raw symmetric key.

        Returns:
             the key from the `k` parameter, base64u-decoded
        """
        return self.cryptography_key  # type: ignore

    def encrypt(
        self,
        plaintext: bytes,
        *,
        aad: Optional[bytes] = None,
        alg: Optional[str] = None,
        iv: Optional[bytes] = None,
    ) -> Tuple[BinaPy, BinaPy, BinaPy]:
        """Encrypt arbitrary data using this key. Supports Authenticated Encryption with the Additional Authenticated Data (`aad`). An Initializatin Vector (IV) will be generated automatically. You can choose your own IV by providing the `iv` parameter (only use this if you know what you are doing).

        This return the ciphertext, the authentication tag, and the used IV (if an IV was provided as parameter, the same IV is returned).

        Args:
          plaintext: the plaintext to encrypt
          aad: the Additional Authentication Data, if any
          alg: the encryption alg to use
          iv: the IV to use, if you want a specific value

        Returns:
            a (ciphertext, authentication_tag, iv) tuple
        """
        encalg = select_alg(self.alg, alg, self.ENCRYPTION_ALGORITHMS)

        if iv is None:
            iv = encalg.generate_iv()

        wrapper: BaseAESEncryptionAlg = encalg(self.cryptography_key)
        ciphertext, tag = wrapper.encrypt(plaintext, iv=iv, aad=aad)
        return ciphertext, BinaPy(iv), tag

    @property
    def key_size(self) -> int:
        """The key size, in bits.

        Returns:
            the key size in bits
        """
        return len(self.key) * 8

    def decrypt(
        self,
        ciphertext: bytes,
        *,
        iv: bytes,
        tag: bytes,
        aad: Optional[bytes] = None,
        alg: Optional[str] = None,
    ) -> BinaPy:
        """Decrypt arbitrary data.

        Args:
          ciphertext: the encrypted data
          iv: the Initialization Vector (must be the same as used during encryption)
          tag: the authentication tag
          aad: the Additional Authenticated Data (must be the same as used during encryption)
          alg: the decryption alg (must be the same as used during encryption)

        Returns:
            the decrypted clear-text
        """
        encalg = select_alg(self.alg, alg, self.ENCRYPTION_ALGORITHMS)
        decryptor: BaseAESEncryptionAlg = encalg(self.cryptography_key)
        plaintext: bytes = decryptor.decrypt(ciphertext, auth_tag=tag, iv=iv, aad=aad)

        return BinaPy(plaintext)

    def supported_key_management_algorithms(self) -> List[str]:
        """Return the list of supported Key Management algorithms, usable for key (un)wrapping with this key.

        Returns:
            a list of supported algorithms identifiers
        """
        return [
            name
            for name, alg in self.KEY_MANAGEMENT_ALGORITHMS.items()
            if alg.supports_key(self.cryptography_key)  # type: ignore
        ]

    def supported_encryption_algorithms(self) -> List[str]:
        """Return the list of supported Encryption/Decryption algorithms with this key.

        Returns:
            a list of supported algorithms identifiers
        """
        return [
            name
            for name, alg in self.ENCRYPTION_ALGORITHMS.items()
            if alg.supports_key(self.cryptography_key)
        ]



"""This module implements the JWS Compact format."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional, Tuple, Union

from backports.cached_property import cached_property
from binapy import BinaPy

from jwskate.jwk.base import Jwk
from jwskate.token import BaseCompactToken

from .signature import JwsSignature

if TYPE_CHECKING:
    from .json import JwsJsonFlat, JwsJsonGeneral  # pragma: no cover


class InvalidJws(ValueError):
    """Raised when an invalid Jws is parsed."""


class JwsCompact(BaseCompactToken):
    """Represents a Json Web Signature (JWS), using compact serialization, as defined in RFC7515."""

    def __init__(self, value: Union[bytes, str]):
        """Initialize a Jws, from its compact representation.

        Args:
            value: the JWS token value
        """
        super().__init__(value)

        header, payload, signature = self.split(self.value)

        try:
            self.headers = BinaPy(header).decode_from("b64u").parse_from("json")
        except ValueError:
            raise InvalidJws(
                "Invalid JWS header: it must be a Base64URL-encoded JSON object"
            )

        try:
            self.payload = BinaPy(payload).decode_from("b64u")
        except ValueError:
            raise InvalidJws(
                "Invalid JWS payload: it must be a Base64URL-encoded binary data (bytes)"
            )

        try:
            self.signature = BinaPy(signature).decode_from("b64u")
        except ValueError:
            raise InvalidJws(
                "Invalid JWS signature: it must be a Base64URL-encoded binary data (bytes)"
            )

    @classmethod
    def split(cls, value: bytes) -> Tuple[BinaPy, BinaPy, BinaPy]:
        """Splits a JWS token value into its (header, payload, signature) parts.

        Args:
          value: the JWS token value

        Returns:
            a (header, payload, signature)

        Raises:
            InvalidJws: if the provided value doesn't have 2 dots.
        """
        if value.count(b".") != 2:
            raise InvalidJws(
                "A JWS must contain a header, a payload and a signature, separated by dots"
            )

        header, payload, signature = value.split(b".")
        return BinaPy(header), BinaPy(payload), BinaPy(signature)

    @classmethod
    def sign(
        cls,
        payload: bytes,
        jwk: Union[Jwk, Dict[str, Any]],
        alg: Optional[str] = None,
        extra_headers: Optional[Dict[str, Any]] = None,
    ) -> "JwsCompact":
        """Sign a payload and returns the resulting JwsCompact.

        Args:
          payload: the payload to sign
          jwk: the jwk to use to sign this payload
          alg: the alg to use
          extra_headers: additional headers to add to the Jws Headers

        Returns:
          the resulting token
        """
        jwk = Jwk(jwk)

        headers = dict(extra_headers or {}, alg=alg)
        kid = jwk.get("kid")
        if kid:
            headers["kid"] = kid

        signed_part = JwsSignature.assemble_signed_part(headers, payload)
        signature = jwk.sign(signed_part, alg=alg)
        return cls.from_parts(signed_part, signature)

    @classmethod
    def from_parts(
        cls, signed_part: Union[bytes, str], signature: Union[bytes, str]
    ) -> "JwsCompact":
        """Constructs a JWS token based on its signed part and signature values.

        Signed part is the concatenation of the header and payload, both encoded in Base64-Url, and joined by a dot.

        Args:
          signed_part: the signed part
          signature: the signature value

        Returns:
            the resulting token
        """
        if not isinstance(signed_part, bytes):
            signed_part = signed_part.encode("ascii")

        return cls(b".".join((signed_part, BinaPy(signature).to("b64u"))))

    @cached_property
    def signed_part(self) -> bytes:
        """Returns the signed part (header + payload) from this JwsCompact.

        Returns:
            the signed part
        """
        return b".".join(self.value.split(b".", 2)[:2])

    @cached_property
    def alg(self) -> str:
        """Get the signature algorithm (alg) from this token headers.

        Returns:
            the `alg` value
        Raises:
            AttributeError: if the `alg` header value is not a string
        """
        alg = self.get_header("alg")
        if alg is None or not isinstance(alg, str):
            raise AttributeError("This JWS doesn't have a valid 'alg' header")
        return alg

    def verify_signature(
        self,
        jwk: Union[Jwk, Dict[str, Any]],
        *,
        alg: Optional[str] = None,
        algs: Optional[Iterable[str]] = None,
    ) -> bool:
        """Verify the signature from this JwsCompact using a Jwk.

        Args:
          jwk: the Jwk to use to validate this signature
          alg: the alg to use, if there is only 1 allowed
          algs: the allowed algs, if here are several

        Returns:
         `True` if the signature matches, `False` otherwise
        """
        jwk = Jwk(jwk)
        return jwk.verify(self.signed_part, self.signature, alg=alg, algs=algs)

    def flat_json(self, unprotected_header: Any = None) -> JwsJsonFlat:
        """Create a JWS in JSON flat format based on this Compact JWS.

        Args:
          unprotected_header: optional unprotected header to include in the JWS JSON

        Returns:
            the resulting token
        """
        from .json import JwsJsonFlat

        protected, payload, signature = self.split(self.value)
        content = {
            "payload": payload.ascii(),
            "protected": protected.ascii(),
            "signature": signature.ascii(),
        }
        if unprotected_header is not None:
            content["header"] = unprotected_header
        return JwsJsonFlat(content)

    def general_json(self, unprotected_header: Any = None) -> JwsJsonGeneral:
        """Create a JWS in JSON General format based on this JWS Compact.

        The resulting token will have a single signature which is the one from this token.

        Args:
            unprotected_header: optional unprotected header to include in the JWS JSON

        Returns:
            the resulting token
        """
        jws = self.flat_json(unprotected_header)
        return jws.generalize()



import pytest
from cryptography.hazmat.primitives.asymmetric import rsa

from jwskate import InvalidJwk, Jwk, RSAJwk


def test_jwk_copy() -> None:
    jwk1 = Jwk.generate_for_kty("RSA")

    jwk2 = Jwk(jwk1)
    assert jwk1 is not jwk2

    jwk3 = jwk1.copy()
    assert jwk1 == jwk3
    assert jwk1 is not jwk3


def test_invalid_jwk() -> None:
    with pytest.raises(ValueError):
        # kty is not str
        Jwk({"kty": 1.5})

    with pytest.raises(ValueError):
        # kty is unknown
        Jwk({"kty": "caesar13"})

    with pytest.raises(InvalidJwk):
        # attributes are missing
        Jwk({"kty": "RSA"})

    with pytest.raises(InvalidJwk):
        # x is not a base64u
        Jwk({"kty": "RSA", "x": "$+!"})

    with pytest.raises(InvalidJwk):
        # attribute 'd' (private exponent) is missing
        Jwk(
            {
                "kty": "RSA",
                "n": "oRHn4oGv23ylRL3RSsL4p_e6Ywinnj2N2tT5OLe5pEZTg-LFBhjFxcJaB-p1dh6XX47EtSfa-JHffU0o5ZRK2ySyNDtlrFAkOpAHH6U83ayE2QPYGzrFrrvHDa8wIMUWymzxpPwGgKBwZZqtTT6d-iy4Ux3AWV-bUv6Z7WijHnOy7aVzZ4dFERLVf2FaaYXDET7GO4v-oQ5ss_guYdmewN039jxkjz_KrA-0Fyhalf9hL8IHfpdpSlHosrmjORG5y9LkYK0J6zxSBF5ZvLIBK33BTzPPiCMwKLyAcV6qdcAcvV4kthKO0iUKBK4eE8D0N8HcSPvA9F_PpLS_k5F2lw",
                "e": "AQAB",
                "p": "0mzP9sbFxU5YxNNLgUEdRQSO-ojqWrzbI02PfQLGyzXumvOh_Qr73OpHStU8CAAcUBaQdRGidsVdb5cq6JG2zvbEEYiX-dCHqTJs8wfktGCL7eV-ZVh7fhJ1sYVBN20yv8aSH63uUPZnJXR1AUyrvRumuerdPxp8X951PESrJd0",
            }
        )

    with pytest.raises(InvalidJwk):
        # k is not a str
        Jwk({"kty": "oct", "k": 1.23})

    with pytest.raises(InvalidJwk):
        # k is not a base64u
        Jwk({"kty": "oct", "k": "Foo****"})

    with pytest.raises(InvalidJwk):
        # oth is unsupported
        Jwk({"kty": "RSA", "oth": "foo"})

    with pytest.raises(InvalidJwk):
        # key is public and has key_ops: ["sign"]
        Jwk(
            {
                "kty": "EC",
                "key_ops": ["sign"],
                "crv": "P-256",
                "x": "vGVh-60pT34a0JLeiaers66I0JLRilpf5tbnZsa-q3U",
                "y": "y99gwPgQH1lrIBQPwgJoHCoeQjF96M7XfxGXu_Pjyzk",
            }
        )


def test_json() -> None:
    j = '{"kty": "RSA", "n": "5nHd3aefARenRFQn-wVrjBLS-5A0uUiHWfOUt8EpjwE3wADAvVPKLbvRQJfugyrn_RpnLqqZFkwYrVD_u1Uzl9J17XJG75jGjCf-gVs1t9FPpgEsJGYK4RK2_f40AxAc6hKomB9q6_dqIxChDxVCrIrrWd9kRk0T86d8Ade3J4f_iMbremm3woSwI6QD056DkRtAD_v2PZQbUBgSru-PsrJ5l_pxxlGPxzAM4_XH8VfogXI8pWv2UDE1IguVeh371ESCbQbJ7SX2jgNzcvvZMMWs0syfF7P0BzGrh_ONsRcxmtjZgtcOA0TCu2-v8qx7GisgqOWOrzWs7ej5RUsu1sxtT53JG2Y3lrPrgajXTB56mSUaL9ivxEfUD17X_cUznGDNoVqcRdfa27rCtWqd8gL-C7M9bYYgcfpCRPllRvGmWP9oarrG4XoIO17QuhZ5tAoz8oFLM9o6pzR2CeDvmSqFbbTHXYdcpCuvYukIimZP6RruMU9O9YQjgCEGWx06WoTnDqWWjbrId8VqP0xJ_6w0j6av3EWGKLETBbaYRXys4OOy-JZRRHydg-es4tkir4xkMvIG8plxoz_mZbTyO9GA5tMHWzbQciUQFf95Gpiwsa5RDdGZx-guBAN56mtKnUzVG_PmUJ8-pzTATkjVpThBRLWaVPFi0eWLEc2NbF8", "e": "AQAB", "d": "QD----2gB03w9LwB6Zq5QXR-HmU2TfJKyml_LTxNufE8-es70Y1q8RVMYhX6CdgMliSmX_xwcicFOk06_5_hopKrDmiuHl8Z_DmqW6Zyc62H52CsTfmlTttI6cfV-ISlix1opAEebmusy9n3DZ1_2VtiAsylyHP0_BnWGS1rgzoHpLrRLHANK86SzV0NNBMd23cdhVdacBC7DLmpe9yO0-OuwxCwZ_qVe5OKBISiGkfIK08OWZCgO_t8NOHF6yYw8B2pv3wFEHllJHMR3R0akN6MMLmcIZ-qidbvXBgriRROYD3VZwVOgtrI8DlTdPVh3QnpMS3PekoczLgHx6oOhI0ytDvoQ0phZd1D4OJKpPtObBcaDiLcKqPJaV7cYBwHwV2D-DvMkvg9LG1fTZoCUZzVU8NBef8nX8tp3e5qJYYgusdexchdyDwxOigCKQcCSbFYk6cO0vy_XxwGGbqCKeaOxsC-laHR_wEIhrp_NNwVYfDVhvvj70TVNCvA5IWdIiNPaPYSH3klvNwdv6GRXn2RwYwReozgV9wG2dhaA0Sff0WXSgPRw_vF_cLoNIr59bS9X9jxp1-NcU2tei-VxC2g9U1RMFB09hbcXy8pBHoOsxA5UyDTDscZLFgxOiTr_D2GQBRvJch364p0RT0BHMak1Zkfq0LobAB9YWLwCAE", "p": "_AdO8_Y-mLzSnm65JI7Sleoxtf1Ex4LaJuNAzSlUdd5LQeUE-aAQ5qfdP_fjGFwY9uu2EQQ0n8jKKAxWT2z2sj4hEwPoT93SU-pZBA0zm0xeIo9QYPDgIw1ejPnAIdGzuW_F0Oh1ELdZ3XQGskNFywZXjoo4WtCbcwn5sLGlhZRfCFJ9dtJFB3OP3SbxKXWyWOcG6KIB2VivzH5NLz8-9sxiv4mJJj7AMvWA9gt0Qo1Alz_O6Xeg3iiV4qIzf0ioPJ7iy6UCVWpohdcwk2Cb6DHswICqlcOtKKRujLGhxtKJ98woSe-v76PuhZ7MWmS6GuF3iRud4AfsZGZWbue1oQ", "q": "6hN9GJygBBxjQxTLkZJZK3yS2GLJCWjNb5Egf6sWz3J-xvXBmBAszgC3C-OM7EMHr537nzxmLEN-wmIvx-U0OQrzqa5M9zCPdR2SMncRyQdw-nEBLcSqh9gocERjvdROrbZsCz_iLltNoCwCDlZQVE6HJkoVZ5AhAVhj50sQ9jlgWyhYnZMdDdKUfaJpfoADlsYmI49UZKCg2H5yJq1fws_Zh9Mi1z-we7rupOGp8rzgoQkzv68ljbY1yG-TF2Z_W99I32QseNzBRjFp99yXTgp3YcJxNtFnDBdvHLNnRHBwASlD_ZbDxg2p2xILVB_bUHZFW0W3CBzRNFi2bv3h_w", "dp": "aLsYwiSYCpyc4Z2dbmWzePzjP39J76aexP421YrRQFHp8C4djSZJH7CuLoDybBMJhMKa3CNlQukLqOzHiSX8tkE_OUmsZlQFrT17VEWwJl7r12y6uC4g1jAeFHNMtkEQcITULWYMD7BBtdcbWUS_Ygj2pZMmrAZ4Mqv4iMapxALOIwU0ggYLDXemVv5xxQrV3D_VDSMVpZ5HH7F0naeooKJ6fqHGzo_RCtwehSBpZaaRKsknULmXrfornwxMXh5xWw-jq4CcoaYgXU35L6U75JeqjKxrNuUjtfnuvqSqV5byInlCXMcv02PKINjGjuHAvJ7pL568Una4c1hbnqbHQQ", "dq": "RcEtBEqYfOEgy3rE90qPfCARep5lnoI2xkqPTrxjfcp28T-HQ5N-Zp1b7xUOh9Gp1rHTrC5JnGM4wSCVcJJjL6SN3EDu-rLj7Vi0molVKX0oM9m9KjBzSSwnUN1wg79i-u1j4S5Wbs4Soeq7ah5areUA7W4iVsxiqY33p5N9KIMMrd2mGr8eZ2Ibkhz2JxZq-2FtOCecVKhxhlKYHeKIqPtbrdhDh7WZGCYqu8Pr60RSBGtDmpnNLR_hgyuMv-pxhaVSiA_IGPRgPFS5aX25MS55SQ6ywk1A0h-howHrgj-ngREVC9sD2F92AKyt55Hev2mfXYW295nu1hShuQ27bQ", "qi": "ZzWzSSYiJHyLRpqUw-GDluHaIYgDV1w70yYtwl222tvJt0TCQkcZmc3tWmC6qYu_7UfFfVDoCVwBu592p6IC0ZD39_eilHaKbDMZ3dOIS9n9h4yfJvvY-4Jd7o5i_LyK5RktvdyZARKrKfL3mYlWCmHC7yYZC88hDh5qkehRT1d52QBKo928mmrkgJZcuuzEVTygTrnCiiFzcd6A7o8wLbtJPBg4WY793xLipiuSEZ8aWQ50hO98MauBO5MJl_C3kZOkEKiq8JYTU-cUHO6kMqlQ866MOccBsco__frxA8yZlZrfMIDOql8z6oS5tpxR5O_acl9fvJ_thwUAVTk3Ow", "kid": "client_assertion_key"}'
    jwk = Jwk.from_json(j)
    assert jwk.public_jwk().as_jwks().to_json() == (
        '{"keys": [{"kty": "RSA", "kid": "client_assertion_key", "n": '
        '"5nHd3aefARenRFQn-wVrjBLS-5A0uUiHWfOUt8EpjwE3wADAvVPKLbvRQJfugyrn_RpnLqqZFkwYrVD_u1Uzl9J17XJG75jGjCf-gVs1t9FPpgEsJGYK4RK2_f40AxAc6hKomB9q6_dqIxChDxVCrIrrWd9kRk0T86d8Ade3J4f_iMbremm3woSwI6QD056DkRtAD_v2PZQbUBgSru-PsrJ5l_pxxlGPxzAM4_XH8VfogXI8pWv2UDE1IguVeh371ESCbQbJ7SX2jgNzcvvZMMWs0syfF7P0BzGrh_ONsRcxmtjZgtcOA0TCu2-v8qx7GisgqOWOrzWs7ej5RUsu1sxtT53JG2Y3lrPrgajXTB56mSUaL9ivxEfUD17X_cUznGDNoVqcRdfa27rCtWqd8gL-C7M9bYYgcfpCRPllRvGmWP9oarrG4XoIO17QuhZ5tAoz8oFLM9o6pzR2CeDvmSqFbbTHXYdcpCuvYukIimZP6RruMU9O9YQjgCEGWx06WoTnDqWWjbrId8VqP0xJ_6w0j6av3EWGKLETBbaYRXys4OOy-JZRRHydg-es4tkir4xkMvIG8plxoz_mZbTyO9GA5tMHWzbQciUQFf95Gpiwsa5RDdGZx-guBAN56mtKnUzVG_PmUJ8-pzTATkjVpThBRLWaVPFi0eWLEc2NbF8", '
        '"e": "AQAB"}]}'
    )


def test_init_from_cryptography() -> None:
    private_key: rsa.RSAPrivateKey = rsa.generate_private_key(65537, 2048)
    jwk = Jwk(private_key)
    assert jwk.kty == "RSA"


def test_init_from_json() -> None:
    j = '{"kty": "RSA", "n": "5nHd3aefARenRFQn-wVrjBLS-5A0uUiHWfOUt8EpjwE3wADAvVPKLbvRQJfugyrn_RpnLqqZFkwYrVD_u1Uzl9J17XJG75jGjCf-gVs1t9FPpgEsJGYK4RK2_f40AxAc6hKomB9q6_dqIxChDxVCrIrrWd9kRk0T86d8Ade3J4f_iMbremm3woSwI6QD056DkRtAD_v2PZQbUBgSru-PsrJ5l_pxxlGPxzAM4_XH8VfogXI8pWv2UDE1IguVeh371ESCbQbJ7SX2jgNzcvvZMMWs0syfF7P0BzGrh_ONsRcxmtjZgtcOA0TCu2-v8qx7GisgqOWOrzWs7ej5RUsu1sxtT53JG2Y3lrPrgajXTB56mSUaL9ivxEfUD17X_cUznGDNoVqcRdfa27rCtWqd8gL-C7M9bYYgcfpCRPllRvGmWP9oarrG4XoIO17QuhZ5tAoz8oFLM9o6pzR2CeDvmSqFbbTHXYdcpCuvYukIimZP6RruMU9O9YQjgCEGWx06WoTnDqWWjbrId8VqP0xJ_6w0j6av3EWGKLETBbaYRXys4OOy-JZRRHydg-es4tkir4xkMvIG8plxoz_mZbTyO9GA5tMHWzbQciUQFf95Gpiwsa5RDdGZx-guBAN56mtKnUzVG_PmUJ8-pzTATkjVpThBRLWaVPFi0eWLEc2NbF8", "e": "AQAB", "d": "QD----2gB03w9LwB6Zq5QXR-HmU2TfJKyml_LTxNufE8-es70Y1q8RVMYhX6CdgMliSmX_xwcicFOk06_5_hopKrDmiuHl8Z_DmqW6Zyc62H52CsTfmlTttI6cfV-ISlix1opAEebmusy9n3DZ1_2VtiAsylyHP0_BnWGS1rgzoHpLrRLHANK86SzV0NNBMd23cdhVdacBC7DLmpe9yO0-OuwxCwZ_qVe5OKBISiGkfIK08OWZCgO_t8NOHF6yYw8B2pv3wFEHllJHMR3R0akN6MMLmcIZ-qidbvXBgriRROYD3VZwVOgtrI8DlTdPVh3QnpMS3PekoczLgHx6oOhI0ytDvoQ0phZd1D4OJKpPtObBcaDiLcKqPJaV7cYBwHwV2D-DvMkvg9LG1fTZoCUZzVU8NBef8nX8tp3e5qJYYgusdexchdyDwxOigCKQcCSbFYk6cO0vy_XxwGGbqCKeaOxsC-laHR_wEIhrp_NNwVYfDVhvvj70TVNCvA5IWdIiNPaPYSH3klvNwdv6GRXn2RwYwReozgV9wG2dhaA0Sff0WXSgPRw_vF_cLoNIr59bS9X9jxp1-NcU2tei-VxC2g9U1RMFB09hbcXy8pBHoOsxA5UyDTDscZLFgxOiTr_D2GQBRvJch364p0RT0BHMak1Zkfq0LobAB9YWLwCAE", "p": "_AdO8_Y-mLzSnm65JI7Sleoxtf1Ex4LaJuNAzSlUdd5LQeUE-aAQ5qfdP_fjGFwY9uu2EQQ0n8jKKAxWT2z2sj4hEwPoT93SU-pZBA0zm0xeIo9QYPDgIw1ejPnAIdGzuW_F0Oh1ELdZ3XQGskNFywZXjoo4WtCbcwn5sLGlhZRfCFJ9dtJFB3OP3SbxKXWyWOcG6KIB2VivzH5NLz8-9sxiv4mJJj7AMvWA9gt0Qo1Alz_O6Xeg3iiV4qIzf0ioPJ7iy6UCVWpohdcwk2Cb6DHswICqlcOtKKRujLGhxtKJ98woSe-v76PuhZ7MWmS6GuF3iRud4AfsZGZWbue1oQ", "q": "6hN9GJygBBxjQxTLkZJZK3yS2GLJCWjNb5Egf6sWz3J-xvXBmBAszgC3C-OM7EMHr537nzxmLEN-wmIvx-U0OQrzqa5M9zCPdR2SMncRyQdw-nEBLcSqh9gocERjvdROrbZsCz_iLltNoCwCDlZQVE6HJkoVZ5AhAVhj50sQ9jlgWyhYnZMdDdKUfaJpfoADlsYmI49UZKCg2H5yJq1fws_Zh9Mi1z-we7rupOGp8rzgoQkzv68ljbY1yG-TF2Z_W99I32QseNzBRjFp99yXTgp3YcJxNtFnDBdvHLNnRHBwASlD_ZbDxg2p2xILVB_bUHZFW0W3CBzRNFi2bv3h_w", "dp": "aLsYwiSYCpyc4Z2dbmWzePzjP39J76aexP421YrRQFHp8C4djSZJH7CuLoDybBMJhMKa3CNlQukLqOzHiSX8tkE_OUmsZlQFrT17VEWwJl7r12y6uC4g1jAeFHNMtkEQcITULWYMD7BBtdcbWUS_Ygj2pZMmrAZ4Mqv4iMapxALOIwU0ggYLDXemVv5xxQrV3D_VDSMVpZ5HH7F0naeooKJ6fqHGzo_RCtwehSBpZaaRKsknULmXrfornwxMXh5xWw-jq4CcoaYgXU35L6U75JeqjKxrNuUjtfnuvqSqV5byInlCXMcv02PKINjGjuHAvJ7pL568Una4c1hbnqbHQQ", "dq": "RcEtBEqYfOEgy3rE90qPfCARep5lnoI2xkqPTrxjfcp28T-HQ5N-Zp1b7xUOh9Gp1rHTrC5JnGM4wSCVcJJjL6SN3EDu-rLj7Vi0molVKX0oM9m9KjBzSSwnUN1wg79i-u1j4S5Wbs4Soeq7ah5areUA7W4iVsxiqY33p5N9KIMMrd2mGr8eZ2Ibkhz2JxZq-2FtOCecVKhxhlKYHeKIqPtbrdhDh7WZGCYqu8Pr60RSBGtDmpnNLR_hgyuMv-pxhaVSiA_IGPRgPFS5aX25MS55SQ6ywk1A0h-howHrgj-ngREVC9sD2F92AKyt55Hev2mfXYW295nu1hShuQ27bQ", "qi": "ZzWzSSYiJHyLRpqUw-GDluHaIYgDV1w70yYtwl222tvJt0TCQkcZmc3tWmC6qYu_7UfFfVDoCVwBu592p6IC0ZD39_eilHaKbDMZ3dOIS9n9h4yfJvvY-4Jd7o5i_LyK5RktvdyZARKrKfL3mYlWCmHC7yYZC88hDh5qkehRT1d52QBKo928mmrkgJZcuuzEVTygTrnCiiFzcd6A7o8wLbtJPBg4WY793xLipiuSEZ8aWQ50hO98MauBO5MJl_C3kZOkEKiq8JYTU-cUHO6kMqlQ866MOccBsco__frxA8yZlZrfMIDOql8z6oS5tpxR5O_acl9fvJ_thwUAVTk3Ow", "kid": "client_assertion_key"}'
    jwk = Jwk(j)
    assert jwk.public_jwk().as_jwks().to_json() == (
        '{"keys": [{"kty": "RSA", "kid": "client_assertion_key", "n": '
        '"5nHd3aefARenRFQn-wVrjBLS-5A0uUiHWfOUt8EpjwE3wADAvVPKLbvRQJfugyrn_RpnLqqZFkwYrVD_u1Uzl9J17XJG75jGjCf-gVs1t9FPpgEsJGYK4RK2_f40AxAc6hKomB9q6_dqIxChDxVCrIrrWd9kRk0T86d8Ade3J4f_iMbremm3woSwI6QD056DkRtAD_v2PZQbUBgSru-PsrJ5l_pxxlGPxzAM4_XH8VfogXI8pWv2UDE1IguVeh371ESCbQbJ7SX2jgNzcvvZMMWs0syfF7P0BzGrh_ONsRcxmtjZgtcOA0TCu2-v8qx7GisgqOWOrzWs7ej5RUsu1sxtT53JG2Y3lrPrgajXTB56mSUaL9ivxEfUD17X_cUznGDNoVqcRdfa27rCtWqd8gL-C7M9bYYgcfpCRPllRvGmWP9oarrG4XoIO17QuhZ5tAoz8oFLM9o6pzR2CeDvmSqFbbTHXYdcpCuvYukIimZP6RruMU9O9YQjgCEGWx06WoTnDqWWjbrId8VqP0xJ_6w0j6av3EWGKLETBbaYRXys4OOy-JZRRHydg-es4tkir4xkMvIG8plxoz_mZbTyO9GA5tMHWzbQciUQFf95Gpiwsa5RDdGZx-guBAN56mtKnUzVG_PmUJ8-pzTATkjVpThBRLWaVPFi0eWLEc2NbF8", '
        '"e": "AQAB"}]}'
    )


def test_missing_kty() -> None:
    with pytest.raises(ValueError):
        Jwk({"foo": "kty_is_missing"})


def test_include_kid() -> None:
    jwk = Jwk({"kty": "oct", "k": "foobar"}, include_kid_thumbprint=True)
    assert jwk.kid == jwk.thumbprint() == "p91YUDdd513xDMIoKEJAySww4jB3hionP-1CUWx6b8g"


def test_getattr() -> None:
    jwk = Jwk.generate_for_kty("oct")
    with pytest.raises(AttributeError):
        jwk.foo


def test_setattr() -> None:
    jwk = Jwk.generate_for_kty("oct")
    with pytest.raises(RuntimeError):
        jwk["k"] = "foo"


def test_invalid_alg() -> None:
    jwk = Jwk({"kty": "oct", "k": "foobar", "alg": 1.34})
    with pytest.raises(TypeError):
        jwk.alg


def test_invalid_class_for_kty() -> None:
    with pytest.raises(TypeError):
        RSAJwk({"kty": "oct", "k": "foobar"})


@pytest.mark.parametrize(
    "private_key_ops, public_key_ops",
    [
        ("sign", "verify"),
        ("decrypt", "encrypt"),
        ("unwrapKey", "wrapKey"),
    ],
)
def test_key_ops(private_key_ops: str, public_key_ops: str) -> None:
    private_jwk = Jwk.generate_for_kty("EC", key_ops=[private_key_ops])
    public_jwk = private_jwk.public_jwk()
    assert public_key_ops in public_jwk.key_ops
    assert private_key_ops not in public_jwk.key_ops



"""This module implements Elliptic Curve Diffie-Hellman based Key Management algorithms."""

from typing import Any, Type, Union

from binapy import BinaPy
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import ec, x448, x25519
from cryptography.hazmat.primitives.kdf.concatkdf import ConcatKDFHash

from ..base import BaseAsymmetricAlg, BaseKeyManagementAlg
from .aeskw import A128KW, A192KW, A256KW, BaseAesKeyWrap


class EcdhEs(
    BaseKeyManagementAlg,
    BaseAsymmetricAlg[
        Union[ec.EllipticCurvePrivateKey, x25519.X25519PrivateKey, x448.X448PrivateKey],
        Union[ec.EllipticCurvePublicKey, x25519.X25519PublicKey, x448.X448PublicKey],
    ],
):
    """Elliptic Curve Diffie-Hellman Ephemeral Static key agreement using Concat KDF."""

    name = "ECDH-ES"
    description = __doc__
    public_key_class = (
        ec.EllipticCurvePublicKey,
        x25519.X25519PublicKey,
        x448.X448PublicKey,
    )
    private_key_class = (
        ec.EllipticCurvePrivateKey,
        x25519.X25519PrivateKey,
        x448.X448PrivateKey,
    )

    @classmethod
    def otherinfo(cls, alg: str, apu: bytes, apv: bytes, keysize: int) -> BinaPy:
        """Build the "otherinfo" parameter for Concat KDF Hash.

        Args:
          alg: identifier for the encryption alg
          apu: Agreement PartyUInfo
          apv: Agreement PartyVInfo
          keysize: length of the generated key

        Returns:
            the "otherinfo" value
        """
        algorithm_id = BinaPy.from_int(len(alg), length=4) + BinaPy(alg)
        partyuinfo = BinaPy.from_int(len(apu), length=4) + apu
        partyvinfo = BinaPy.from_int(len(apv), length=4) + apv
        supppubinfo = BinaPy.from_int(keysize or keysize, length=4)
        otherinfo = b"".join((algorithm_id, partyuinfo, partyvinfo, supppubinfo))
        return BinaPy(otherinfo)

    @classmethod
    def ecdh(
        cls,
        private_key: Union[
            ec.EllipticCurvePrivateKey, x25519.X25519PrivateKey, x448.X448PrivateKey
        ],
        public_key: Union[
            ec.EllipticCurvePublicKey, x25519.X25519PublicKey, x448.X448PublicKey
        ],
    ) -> BinaPy:
        """This does an Elliptic Curve Diffie Hellman key exchange.

        This derives a shared key between a sender and a receiver, based on a public and a private key from each side.
        ECDH exchange produces the same key with either a sender private key and a recipient public key,
        or the matching sender public key and recipient private key.

        Args:
          private_key: a private EC key
          public_key: a public EC key

        Returns:
          a shared key
        """
        if isinstance(private_key, ec.EllipticCurvePrivateKey) and isinstance(
            public_key, ec.EllipticCurvePublicKey
        ):
            shared_key = private_key.exchange(ec.ECDH(), public_key)
        elif isinstance(private_key, x25519.X25519PrivateKey) and isinstance(
            public_key, x25519.X25519PublicKey
        ):
            shared_key = private_key.exchange(public_key)
        elif isinstance(private_key, x448.X448PrivateKey) and isinstance(
            public_key, x448.X448PublicKey
        ):
            shared_key = private_key.exchange(public_key)
        else:
            raise ValueError(
                "Invalid or unsupported private/public key combination for ECDH",
                private_key,
                public_key,
            )
        return BinaPy(shared_key)

    @classmethod
    def derive(
        cls,
        *,
        private_key: Union[
            ec.EllipticCurvePrivateKey, x25519.X25519PrivateKey, x448.X448PrivateKey
        ],
        public_key: Union[
            ec.EllipticCurvePublicKey, x25519.X25519PublicKey, x448.X448PublicKey
        ],
        otherinfo: bytes,
        key_size: int,
    ) -> BinaPy:
        """Derive a key using ECDH and Concat KDF Hash.

        Args:
          private_key: the private key
          public_key: the public key
          otherinfo: the Concat KDF "otherinfo" parameter
          key_size: the expected CEK key size

        Returns:
            the derived key
        """
        shared_key = cls.ecdh(private_key, public_key)
        ckdf = ConcatKDFHash(
            algorithm=hashes.SHA256(), length=key_size // 8, otherinfo=otherinfo
        )
        return BinaPy(ckdf.derive(shared_key))

    def generate_ephemeral_key(
        self,
    ) -> Union[
        ec.EllipticCurvePrivateKey, x25519.X25519PrivateKey, x448.X448PrivateKey
    ]:
        """Generate an ephemeral key that is suitable for use with this algorithm.

        Returns:
            a generated EllipticCurvePrivateKey, on the same curve as this algorithm key
        """
        if isinstance(
            self.key, (ec.EllipticCurvePrivateKey, ec.EllipticCurvePublicKey)
        ):
            return ec.generate_private_key(self.key.curve)
        elif isinstance(self.key, (x25519.X25519PrivateKey, x25519.X25519PublicKey)):
            return x25519.X25519PrivateKey.generate()
        elif isinstance(self.key, (x448.X448PublicKey, x448.X448PrivateKey)):
            return x448.X448PrivateKey.generate()

    def sender_key(
        self,
        ephemeral_private_key: Union[
            ec.EllipticCurvePrivateKey, x25519.X25519PrivateKey, x448.X448PrivateKey
        ],
        *,
        alg: str,
        key_size: int,
        **headers: Any,
    ) -> BinaPy:
        """Compute a CEK for encryption of a message. This method is meant for usage by a sender.

        Args:
          ephemeral_private_key: the EPK to use for this key
          alg: the content encryption algorithm identifier
          key_size: the expected CEK size
          **headers: additional headers to include for CEK derivation

        Returns:
            the CEK for encryption by the sender
        """
        with self.public_key_required() as key:
            apu = BinaPy(headers.get("apu", b"")).decode_from("b64u")
            apv = BinaPy(headers.get("apv", b"")).decode_from("b64u")
            otherinfo = self.otherinfo(alg, apu, apv, key_size)
            cek = self.derive(
                private_key=ephemeral_private_key,
                public_key=key,
                otherinfo=otherinfo,
                key_size=key_size,
            )
            return cek

    def recipient_key(
        self,
        ephemeral_public_key: Union[
            ec.EllipticCurvePublicKey, x25519.X25519PublicKey, x448.X448PublicKey
        ],
        *,
        alg: str,
        key_size: int,
        **headers: Any,
    ) -> BinaPy:
        """Compute a shared key. This method is meant for use by the recipient of an encrypted message.

        Args:
          ephemeral_public_key: the EPK, as received from sender
          alg: the content encryption algorithm identifier
          key_size: the CEK size
          **headers: additional headers as received from sender

        Returns:
            the CEK for decryption by the recipient
        """
        with self.private_key_required() as key:
            apu = BinaPy(headers.get("apu", b"")).decode_from("b64u")
            apv = BinaPy(headers.get("apv", b"")).decode_from("b64u")
            otherinfo = self.otherinfo(alg, apu, apv, key_size)
            cek = self.derive(
                private_key=key,
                public_key=ephemeral_public_key,
                otherinfo=otherinfo,
                key_size=key_size,
            )
            return cek


class BaseEcdhEs_AesKw(EcdhEs):
    """Base class for ECDH-ES+AESKW algorithms."""

    kwalg: Type[BaseAesKeyWrap]

    def wrap_key_with_epk(
        self,
        plainkey: bytes,
        ephemeral_private_key: Union[
            ec.EllipticCurvePrivateKey, x25519.X25519PrivateKey, x448.X448PrivateKey
        ],
        **headers: Any,
    ) -> BinaPy:
        """Wraps a key for content encryption.

        Args:
          plainkey: the key to wrap
          ephemeral_private_key: the EPK to use
          **headers: additional headers for CEK derivation

        Returns:
            the wrapped CEK
        """
        aes_key = self.sender_key(
            ephemeral_private_key, key_size=self.kwalg.key_size, **headers
        )
        return self.kwalg(aes_key).wrap_key(plainkey)

    def unwrap_key_with_epk(
        self,
        cipherkey: bytes,
        ephemeral_public_key: Union[
            ec.EllipticCurvePublicKey, x25519.X25519PublicKey, x448.X448PublicKey
        ],
        **headers: Any,
    ) -> BinaPy:
        """Unwrap a key for content decryption.

        Args:
          cipherkey: the wrapped key
          ephemeral_public_key: the EPK
          **headers: additional headers for CEK derivation

        Returns:
            the unwrapped key
        """
        aes_key = self.recipient_key(
            ephemeral_public_key, key_size=self.kwalg.key_size, **headers
        )
        return self.kwalg(aes_key).unwrap_key(cipherkey)


class EcdhEs_A128KW(BaseEcdhEs_AesKw):
    """ECDH-ES using Concat KDF and "A128KW" wrapping."""

    name = "ECDH-ES+A128KW"
    description = __doc__
    kwalg = A128KW


class EcdhEs_A192KW(BaseEcdhEs_AesKw):
    """ECDH-ES using Concat KDF and "A192KW" wrapping."""

    name = "ECDH-ES+A192KW"
    description = __doc__
    kwalg = A192KW


class EcdhEs_A256KW(BaseEcdhEs_AesKw):
    """ECDH-ES using Concat KDF and "A256KW" wrapping."""

    name = "ECDH-ES+A256KW"
    description = __doc__
    kwalg = A256KW
